import unittest
import astrophot as ap
import numpy as np
import torch


class TestWindow(unittest.TestCase):
    def test_window_creation(self):

        window1 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))

        window1.to(dtype=torch.float64, device="cpu")

        self.assertEqual(window1.origin[0], 0, "Window should store origin")
        self.assertEqual(window1.origin[1], 6, "Window should store origin")
        self.assertEqual(window1.shape[0], 100, "Window should store shape")
        self.assertEqual(window1.shape[1], 110, "Window should store shape")
        self.assertEqual(window1.center[0], 50.0, "Window should determine center")
        self.assertEqual(window1.center[1], 61.0, "Window should determine center")

        self.assertRaises(Exception, ap.image.Window)

        x = str(window1)
        x = repr(window1)

        wcs = window1.get_astropywcs()

    def test_window_crop(self):

        window1 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))

        window1.crop_to_pixel([[10, 90], [15, 105]])
        self.assertTrue(
            np.all(window1.origin.detach().cpu().numpy() == np.array([10.0, 21])),
            "crop pixels should move origin",
        )
        self.assertTrue(
            np.all(window1.pixel_shape.detach().cpu().numpy() == np.array([80, 90])),
            "crop pixels should change shape",
        )

        window2 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))
        window2.crop_pixel((5,))
        self.assertTrue(
            np.all(window2.origin.detach().cpu().numpy() == np.array([5.0, 11.0])),
            "crop pixels should move origin",
        )
        self.assertTrue(
            np.all(window2.pixel_shape.detach().cpu().numpy() == np.array([90, 100])),
            "crop pixels should change shape",
        )
        window2.pad_pixel((5,))

        window2 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))
        window2.crop_pixel((5, 6))
        self.assertTrue(
            np.all(window2.origin.detach().cpu().numpy() == np.array([5.0, 12.0])),
            "crop pixels should move origin",
        )
        self.assertTrue(
            np.all(window2.pixel_shape.detach().cpu().numpy() == np.array([90, 98])),
            "crop pixels should change shape",
        )
        window2.pad_pixel((5, 6))

        window2 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))
        window2.crop_pixel((5, 6, 7, 8))
        self.assertTrue(
            np.all(window2.origin.detach().cpu().numpy() == np.array([5.0, 12.0])),
            "crop pixels should move origin",
        )
        self.assertTrue(
            np.all(window2.pixel_shape.detach().cpu().numpy() == np.array([88, 96])),
            "crop pixels should change shape",
        )
        window2.pad_pixel((5, 6, 7, 8))

        self.assertTrue(
            np.all(window2.origin.detach().cpu().numpy() == np.array([0.0, 6.0])),
            "pad pixels should move origin",
        )
        self.assertTrue(
            np.all(window2.pixel_shape.detach().cpu().numpy() == np.array([100, 110])),
            "pad pixels should change shape",
        )

    def test_window_get_indices(self):

        window1 = ap.image.Window(origin=(0, 6), pixel_shape=(100, 110))
        xstep, ystep = np.meshgrid(range(100), range(110), indexing="xy")
        zstep = xstep + ystep
        window2 = ap.image.Window(origin=(15, 15), pixel_shape=(30, 200))

        zsliced = zstep[window1.get_self_indices(window2)]
        self.assertTrue(
            np.all(zsliced == zstep[9:110, 15:45]),
            "window slices should get correct part of image",
        )
        zsliced = zstep[window2.get_other_indices(window1)]
        self.assertTrue(
            np.all(zsliced == zstep[9:110, 15:45]),
            "window slices should get correct part of image",
        )

    def test_window_arithmetic(self):

        windowbig = ap.image.Window(origin=(0, 0), pixel_shape=(100, 110))
        windowsmall = ap.image.Window(origin=(40, 40), pixel_shape=(20, 30))

        # Logical or, size
        ######################################################################
        big_or_small = windowbig | windowsmall
        self.assertEqual(
            big_or_small.origin[0],
            0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_small.shape[0],
            100,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.origin[0],
            40,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.shape[0],
            20,
            "logical or of images should not affect initial images",
        )

        # Logical and, size
        ######################################################################
        big_and_small = windowbig & windowsmall
        self.assertEqual(
            big_and_small.origin[0],
            40,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_small.shape[0],
            20,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_small.shape[1],
            30,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.origin[0],
            40,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowsmall.shape[0],
            20,
            "logical and of images should not affect initial images",
        )

        # Logical or, offset
        ######################################################################
        windowoffset = ap.image.Window(origin=(40, -20), pixel_shape=(100, 90))
        big_or_offset = windowbig | windowoffset
        self.assertEqual(
            big_or_offset.origin[0],
            0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_offset.origin[1],
            -20,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_offset.shape[0],
            140,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            big_or_offset.shape[1],
            130,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.origin[0],
            40,
            "logical or of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.shape[0],
            100,
            "logical or of images should not affect initial images",
        )

        # Logical and,  offset
        ######################################################################
        big_and_offset = windowbig & windowoffset
        self.assertEqual(
            big_and_offset.origin[0],
            40,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_offset.origin[1],
            0,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_offset.shape[0],
            60,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            big_and_offset.shape[1],
            70,
            "logical and of images should take overlap region",
        )
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.origin[0],
            40,
            "logical and of images should not affect initial images",
        )
        self.assertEqual(
            windowoffset.shape[0],
            100,
            "logical and of images should not affect initial images",
        )

        # Logical ior, size
        ######################################################################
        windowbig |= windowsmall
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.shape[0],
            100,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowsmall.origin[0],
            40,
            "logical or of images should not affect input image",
        )
        self.assertEqual(
            windowsmall.shape[0],
            20,
            "logical or of images should not affect input image",
        )

        # Logical ior, offset
        ######################################################################
        windowbig |= windowoffset
        self.assertEqual(
            windowbig.origin[0],
            0,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.origin[1],
            -20,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.shape[0],
            140,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowbig.shape[1],
            130,
            "logical or of images should take largest bounding box",
        )
        self.assertEqual(
            windowoffset.origin[0],
            40,
            "logical or of images should not affect input image",
        )
        self.assertEqual(
            windowoffset.shape[0],
            100,
            "logical or of images should not affect input image",
        )

        # Logical iand, offset
        ######################################################################
        windowbig = ap.image.Window(origin=(0, 0), pixel_shape=(100, 110))
        windowbig &= windowoffset
        self.assertEqual(
            windowbig.origin[0], 40, "logical and of images should take overlap region"
        )
        self.assertEqual(windowbig.origin[1], 0, "logical and of images should take overlap region")
        self.assertEqual(windowbig.shape[0], 60, "logical and of images should take overlap region")
        self.assertEqual(windowbig.shape[1], 70, "logical and of images should take overlap region")
        self.assertEqual(
            windowoffset.origin[0],
            40,
            "logical and of images should not affect input image",
        )
        self.assertEqual(
            windowoffset.shape[0],
            100,
            "logical and of images should not affect input image",
        )

        windowbig &= windowsmall

        self.assertEqual(
            windowbig,
            windowsmall,
            "logical and of images should take overlap region, equality should be internally determined",
        )

    def test_window_state(self):
        window_init = ap.image.Window(
            origin=[1.0, 2.0],
            pixel_shape=[10, 15],
            pixelscale=1,
            projection="orthographic",
            reference_radec=(0, 0),
        )
        window = ap.image.Window(state=window_init.get_state())
        self.assertEqual(window.origin[0].item(), 1.0, "Window initialization should read state")
        self.assertEqual(window.shape[0].item(), 10.0, "Window initialization should read state")
        self.assertEqual(
            window.pixelscale[0][0].item(),
            1.0,
            "Window initialization should read state",
        )

        state = window.get_state()
        self.assertEqual(
            state["reference_imagexy"][1], 2.0, "Window get state should collect values"
        )
        self.assertEqual(state["pixel_shape"][1], 15.0, "Window get state should collect values")
        self.assertEqual(state["pixelscale"][1][0], 0.0, "Window get state should collect values")
        self.assertEqual(
            state["projection"],
            "orthographic",
            "Window get state should collect values",
        )
        self.assertEqual(
            tuple(state["reference_radec"]),
            (0.0, 0.0),
            "Window get state should collect values",
        )

    def test_window_logic(self):

        window1 = ap.image.Window(origin=[0.0, 1.0], pixel_shape=[10.0, 11.0])
        window2 = ap.image.Window(origin=[0.0, 1.0], pixel_shape=[10.0, 11.0])
        window3 = ap.image.Window(origin=[-0.6, 0.4], pixel_shape=[15.0, 18.0])

        self.assertEqual(window1, window2, "same origin, shape windows should evaluate equal")
        self.assertNotEqual(window1, window3, "Different windows should not evaluate equal")

    def test_window_errors(self):

        # Initialize with conflicting information
        with self.assertRaises(ap.errors.SpecificationConflict):
            window = ap.image.Window(
                origin=[0.0, 1.0], origin_radec=[5.0, 6.0], pixel_shape=[10.0, 11.0]
            )


if __name__ == "__main__":
    unittest.main()



from typing import Optional

import torch
import numpy as np

from ..param import Param_Unlock, Param_SoftLimits, Parameter_Node
from .model_object import Component_Model
from .core_model import AstroPhot_Model
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..image import PSF_Image, Window, Model_Image, Image
from ._shared_methods import select_target
from ..errors import SpecificationConflict

__all__ = ("Point_Source",)


class Point_Source(Component_Model):
    """Describes a point source in the image, this is a delta function at
    some position in the sky. This is typically used to describe
    stars, supernovae, very small galaxies, quasars, asteroids or any
    other object which can essentially be entirely described by a
    position and total flux (no structure).

    """

    model_type = f"point {Component_Model.model_type}"
    parameter_specs = {
        "flux": {"units": "log10(flux)"},
    }
    _parameter_order = Component_Model._parameter_order + ("flux",)
    usable = True

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        if self.psf is None:
            raise ValueError("Point_Source needs psf information")

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        if parameters["flux"].value is not None:
            return
        target_area = target[self.window]
        target_dat = target_area.data.detach().cpu().numpy()
        with Param_Unlock(parameters["flux"]), Param_SoftLimits(parameters["flux"]):
            icenter = target_area.plane_to_pixel(parameters["center"].value)
            edge = np.concatenate(
                (
                    target_dat[:, 0],
                    target_dat[:, -1],
                    target_dat[0, :],
                    target_dat[-1, :],
                )
            )
            edge_average = np.median(edge)
            parameters["flux"].value = np.log10(np.abs(np.sum(target_dat - edge_average)))
            parameters["flux"].uncertainty = torch.std(target_area.data) / (
                np.log(10) * 10 ** parameters["flux"].value
            )

    # Psf convolution should be on by default since this is a delta function
    @property
    def psf_mode(self):
        return "full"

    @psf_mode.setter
    def psf_mode(self, value):
        pass

    def sample(
        self,
        image: Optional[Image] = None,
        window: Optional[Window] = None,
        parameters: Optional[Parameter_Node] = None,
    ):
        """Evaluate the model on the space covered by an image object. This
        function properly calls integration methods and PSF
        convolution. This should not be overloaded except in special
        cases.

        This function is designed to compute the model on a given
        image or within a specified window. It takes care of sub-pixel
        sampling, recursive integration for high curvature regions,
        PSF convolution, and proper alignment of the computed model
        with the original pixel grid. The final model is then added to
        the requested image.

        Args:
          image (Optional[Image]): An AstroPhot Image object (likely a Model_Image)
                                     on which to evaluate the model values. If not
                                     provided, a new Model_Image object will be created.
          window (Optional[Window]): A window within which to evaluate the model.
                                   Should only be used if a subset of the full image
                                   is needed. If not provided, the entire image will
                                   be used.

        Returns:
          Image: The image with the computed model values.

        """
        # Image on which to evaluate model
        if image is None:
            image = self.make_model_image(window=window)

        # Window within which to evaluate model
        if window is None:
            working_window = image.window.copy()
        else:
            working_window = window.copy()

        # Parameters with which to evaluate the model
        if parameters is None:
            parameters = self.parameters

        # Sample the PSF pixels
        if isinstance(self.psf, AstroPhot_Model):
            # Adjust for supersampled PSF
            psf_upscale = torch.round(
                self.psf.target.pixel_length / working_window.pixel_length
            ).int()
            working_window = working_window.rescale_pixel(psf_upscale)
            working_window.shift(-parameters["center"].value)

            # Make the image object to which the samples will be tracked
            working_image = Model_Image(window=working_window)

            # Fill the image using the PSF model
            psf = self.psf(
                image=working_image,
                parameters=parameters[self.psf.name],
            )

            # Scale for point source flux
            working_image.data *= 10 ** parameters["flux"].value

            # Return to original coordinates
            working_image.header.shift(parameters["center"].value)

        elif isinstance(self.psf, PSF_Image):
            psf = self.psf.copy()

            # Adjust for supersampled PSF
            psf_upscale = torch.round(psf.pixel_length / working_window.pixel_length).int()
            working_window = working_window.rescale_pixel(psf_upscale)

            # Make the image object to which the samples will be tracked
            working_image = Model_Image(window=working_window)

            # Compute the center offset
            pixel_center = working_image.plane_to_pixel(parameters["center"].value)
            center_shift = pixel_center - torch.round(pixel_center)
            # working_image.header.pixel_shift(center_shift)
            psf.window.shift(working_image.pixel_to_plane(torch.round(pixel_center)))
            psf.data = self._shift_psf(
                psf=psf.data,
                shift=center_shift,
                shift_method=self.psf_subpixel_shift,
                keep_pad=False,
            )
            psf.data /= torch.sum(psf.data)

            # Scale for psf flux
            psf.data *= 10 ** parameters["flux"].value

            # Fill pixels with the PSF image
            working_image += psf

            # Shift image back to align with original pixel grid
            # working_image.header.pixel_shift(-center_shift)

        else:
            raise SpecificationConflict(
                f"Point_Source must have a psf that is either an AstroPhot_Model or a PSF_Image. not {type(self.psf)}"
            )

        # Return to image pixelscale
        working_image = working_image.reduce(psf_upscale)
        if self.mask is not None:
            working_image.data = working_image.data * torch.logical_not(self.mask)

        # Add the sampled/integrated/convolved pixels to the requested image
        image += working_image

        return image



from types import FunctionType

import torch
import numpy as np

from ..utils.conversions.optimization import (
    boundaries,
    inv_boundaries,
    cyclic_boundaries,
)
from .. import AP_config
from .base import Node
from ..errors import InvalidParameter

__all__ = ["Parameter_Node"]


class Parameter_Node(Node):
    """A node representing parameters and their relative structure.

    The Parameter_Node object stores all information relevant for the
    parameters of a model. At a high level the Parameter_Node
    accomplishes two tasks. The first task is to store the actual
    parameter values, these are represented as pytorch tensors which
    can have any shape; these are leaf nodes. The second task is to
    store the relationship between parameters in a graph structure;
    these are branch nodes. The two tasks are handled by the same type
    of object since there is some overlap between them where a branch
    node acts like a leaf node in certain contexts.

    There are various quantities that a Parameter_Node tracks which
    can be provided as arguments or updated later.

    Args:
      value: The value of a node represents the tensor which will be used by models to compute their projection into the pixels of an image. These can be quite complex, see further down for more details.
      cyclic (bool): Records if the value of a node is cyclic, meaning that if it is updated outside it's limits it should be wrapped back into the limits.
      limits (Tuple[Tensor or None, Tensor or None]): Tracks if a parameter has constraints on the range of values it can take. The first element is the lower limit, the second element is the upper limit. The two elements should either be None (no limit) or tensors with the same shape as the value.
      units (str): The units of the parameter value.
      uncertainty (Tensor or None): represents the uncertainty of the parameter value. This should be None (no uncertainty) or a Tensor with the same shape as the value.
      prof (Tensor or None): This is a profile of values which has no explicit meaning, but can be used to store information which should be kept alongside the value. For example in a spline model the position of the spline points may be a ``prof`` while the flux at each node is the value to be optimized.
      shape (Tuple or None): Can be used to set the shape of the value (number of elements/dimensions). If not provided then the shape will be set by the first time a value is given. Once a shape has been set, if a value is given which cannot be coerced into that shape, then an error will be thrown.

    The ``value`` of a Parameter_Node is somewhat complicated, there
    are a number of states it can take on. The most straightforward is
    just a Tensor, if a Tensor (or just an iterable like a list or
    numpy.ndarray) is provided then the node is required to be a leaf
    node and it will store the value to be accessed later by other
    parts of AstroPhot. Another option is to set the value as another
    node (they will automatically be linked), in this case the node's
    ``value`` is just a wrapper to call for the ``value`` of the
    linked node. Finally, the value may be a function which allows for
    arbitrarily complex values to be computed from other node's
    values. The function must take as an argument the current
    Parameter_Node instance and return a Tensor. Here are some
    examples of the various ways of interacting with the ``value`` for a hypothetical parameter ``P``::

      P.value = 1. # Will create a tensor with value 1.
      P.value = P2 # calling P.value will actually call P2.value
      def compute_value(param):
        return param["P2"].value**2
      P.value = compute_value # calling P.value will call the function as: compute_value(P) which will return P2.value**2

    """

    def __init__(self, name, **kwargs):

        super().__init__(name, **kwargs)
        if "state" in kwargs:
            return
        temp_locked = self.locked
        self.locked = False
        self._value = None
        self.prof = kwargs.get("prof", None)
        self.limits = kwargs.get("limits", [None, None])
        self.cyclic = kwargs.get("cyclic", False)
        self.shape = kwargs.get("shape", None)
        self.value = kwargs.get("value", None)
        self.units = kwargs.get("units", "none")
        self.uncertainty = kwargs.get("uncertainty", None)
        self.to()
        self.locked = temp_locked

    @property
    def value(self):
        """The ``value`` of a Parameter_Node is somewhat complicated, there
        are a number of states it can take on. The most
        straightforward is just a Tensor, if a Tensor (or just an
        iterable like a list or numpy.ndarray) is provided then the
        node is required to be a leaf node and it will store the value
        to be accessed later by other parts of AstroPhot. Another
        option is to set the value as another node (they will
        automatically be linked), in this case the node's ``value`` is
        just a wrapper to call for the ``value`` of the linked
        node. Finally, the value may be a function which allows for
        arbitrarily complex values to be computed from other node's
        values. The function must take as an argument the current
        Parameter_Node instance and return a Tensor. Here are some
        examples of the various ways of interacting with the ``value``
        for a hypothetical parameter ``P``::

          P.value = 1. # Will create a tensor with value 1.
          P.value = P2 # calling P.value will actually call P2.value
          def compute_value(param):
            return param["P2"].value**2
          P.value = compute_value # calling P.value will call the function as: compute_value(P) which will return P2.value**2

        """
        if isinstance(self._value, Parameter_Node):
            return self._value.value
        if isinstance(self._value, FunctionType):
            return self._value(self)

        return self._value

    @property
    def mask(self):
        """The mask tensor is stored internally and it cuts out some values
        from the parameter. This is used by the ``vector`` methods in
        the class to give the parameter DAG a dynamic shape.

        """
        if not self.leaf:
            return self.vector_mask()
        try:
            return self._mask
        except AttributeError:
            return torch.ones(self.shape, dtype=torch.bool, device=AP_config.ap_device)

    @property
    def identities(self):
        """This creates a numpy array of strings which uniquely identify
        every element in the parameter vector. For example a
        ``center`` parameter with two components [x,y] would have
        identities be ``np.array(["123456:0", "123456:1"])`` where the
        first part is the unique id for the Parameter_Node object and
        the second number indexes where in the value tensor it refers
        to.

        """
        if self.leaf:
            idstr = str(self.identity)
            return np.array(tuple(f"{idstr}:{i}" for i in range(self.size)))
        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.identities for node in flat.values())
        if len(vec) > 0:
            return np.concatenate(vec)
        return np.array(())

    @property
    def names(self):
        """Returns a numpy array of names for all the elements of the
        ``vector`` representation where the name is determined by the
        name of the parameters. Note that this does not create a
        unique name for each element and this should only be used for
        graphical purposes on small parameter DAGs.

        """
        if self.leaf:
            S = self.size
            if S == 1:
                return np.array((self.name,))
            return np.array(tuple(f"{self.name}:{i}" for i in range(self.size)))
        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.names for node in flat.values())
        if len(vec) > 0:
            return np.concatenate(vec)
        return np.array(())

    def vector_values(self):
        """The vector representation is for values which correspond to
        fundamental inputs to the parameter DAG. Since the DAG may
        have linked nodes, or functions which produce values derived
        from other node values, the collection of all "values" is not
        necessarily of use for some methods such as fitting
        algorithms. The vector representation is useful for optimizers
        as it gives a fundamental representation of the parameter
        DAG. The vector_values function returns a vector of the
        ``value`` for each leaf node.

        """

        if self.leaf:
            return self.value[self.mask].flatten()

        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.vector_values() for node in flat.values())
        if len(vec) > 0:
            return torch.cat(vec)
        return torch.tensor((), dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    def vector_uncertainty(self):
        """This returns a vector (see vector_values) with the uncertainty for
        each leaf node.

        """
        if self.leaf:
            if self._uncertainty is None:
                self.uncertainty = torch.ones_like(self.value)
            return self.uncertainty[self.mask].flatten()

        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.vector_uncertainty() for node in flat.values())
        if len(vec) > 0:
            return torch.cat(vec)
        return torch.tensor((), dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    def vector_representation(self):
        """This returns a vector (see vector_values) with the representation
        for each leaf node. The representation is an alternative view
        of each value which is mapped into the (-inf, inf) range where
        optimization is more stable.

        """
        return self.vector_transform_val_to_rep(self.vector_values())

    def vector_mask(self):
        """This returns a vector (see vector_values) with the mask for each
        leaf node. Note however that the mask is not itself masked,
        this vector is always the full size of the unmasked parameter
        DAG.

        """
        if self.leaf:
            return self.mask.flatten()

        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.vector_mask() for node in flat.values())
        if len(vec) > 0:
            return torch.cat(vec)
        return torch.tensor((), dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    def vector_identities(self):
        """This returns a vector (see vector_values) with the identities for
        each leaf node.

        """
        if self.leaf:
            return self.identities[self.vector_mask().detach().cpu().numpy()].flatten()
        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.vector_identities() for node in flat.values())
        if len(vec) > 0:
            return np.concatenate(vec)
        return np.array(())

    def vector_names(self):
        """This returns a vector (see vector_values) with the names for each
        leaf node.

        """
        if self.leaf:
            return self.names[self.vector_mask().detach().cpu().numpy()].flatten()
        flat = self.flat(include_locked=False, include_links=False)
        vec = tuple(node.vector_names() for node in flat.values())
        if len(vec) > 0:
            return np.concatenate(vec)
        return np.array(())

    def vector_set_values(self, values):
        """This function allows one to update the full vector of values in a
        single call by providing a tensor of the appropriate size. The
        input will be separated so that the correct elements are
        passed to the correct leaf nodes.

        """
        values = torch.as_tensor(
            values, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        ).flatten()
        if self.leaf:
            self._value[self.mask] = values
            return

        mask = self.vector_mask()
        flat = self.flat(include_locked=False, include_links=False)

        loc = 0
        for node in flat.values():
            node.vector_set_values(
                values[mask[:loc].sum().int() : mask[: loc + node.size].sum().int()]
            )
            loc += node.size

    def vector_set_uncertainty(self, uncertainty):
        """Update the uncertainty vector for this parameter DAG (see
        vector_set_values).

        """
        uncertainty = torch.as_tensor(
            uncertainty, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        if self.leaf:
            if self._uncertainty is None:
                self._uncertainty = torch.ones_like(self.value)
            self._uncertainty[self.mask] = uncertainty
            return

        mask = self.vector_mask()
        flat = self.flat(include_locked=False, include_links=False)

        loc = 0
        for node in flat.values():
            node.vector_set_uncertainty(
                uncertainty[mask[:loc].sum().int() : mask[: loc + node.size].sum().int()]
            )
            loc += node.size

    def vector_set_mask(self, mask):
        """Update the mask vector for this parameter DAG (see
        vector_set_values). Note again that the mask vector is always
        the full size of the DAG.

        """
        mask = torch.as_tensor(mask, dtype=torch.bool, device=AP_config.ap_device)
        if self.leaf:
            self._mask = mask.reshape(self.shape)
            return
        flat = self.flat(include_locked=False, include_links=False)

        loc = 0
        for node in flat.values():
            node.vector_set_mask(mask[loc : loc + node.size])
            loc += node.size

    def vector_set_representation(self, rep):
        """Update the representation vector for this parameter DAG (see
        vector_set_values).

        """
        self.vector_set_values(self.vector_transform_rep_to_val(rep))

    def vector_transform_rep_to_val(self, rep):
        """Used to transform between the ``vector_values`` and
        ``vector_representation`` views of the elements in the DAG
        leafs. This transforms from representation to value.

        The transformation is done based on the limits of each
        parameter leaf. If no limits are provided then the
        representation and value are equivalent. If both are given
        then a ``tan`` and ``arctan`` are used to convert between the
        finite range and the infinite range. If the limits are
        one-sided then the transformation: ``newvalue = value - 1 /
        (value - limit)`` is used.

        """
        rep = torch.as_tensor(rep, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        if self.leaf:
            if self.cyclic:
                val = cyclic_boundaries(rep, (self.limits[0][self.mask], self.limits[1][self.mask]))
            elif self.limits[0] is None and self.limits[1] is None:
                val = rep
            else:
                val = inv_boundaries(
                    rep,
                    (
                        None if self.limits[0] is None else self.limits[0][self.mask],
                        None if self.limits[1] is None else self.limits[1][self.mask],
                    ),
                )
            return val

        mask = self.vector_mask()
        flat = self.flat(include_locked=False, include_links=False)

        loc = 0
        vals = []
        for node in flat.values():
            vals.append(
                node.vector_transform_rep_to_val(
                    rep[mask[:loc].sum().int() : mask[: loc + node.size].sum().int()]
                )
            )
            loc += node.size
        if len(vals) > 0:
            return torch.cat(vals)
        return torch.tensor((), dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    def vector_transform_val_to_rep(self, val):
        """Used to transform between the ``vector_values`` and
        ``vector_representation`` views of the elements in the DAG
        leafs. This transforms from value to representation.

        The transformation is done based on the limits of each
        parameter leaf. If no limits are provided then the
        representation and value are equivalent. If both are given
        then a ``tan`` and ``arctan`` are used to convert between the
        finite range and the infinite range. If the limits are
        one-sided then the transformation: ``newvalue = value - 1 /
        (value - limit)`` is used.

        """
        val = torch.as_tensor(val, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        if self.leaf:
            if self.cyclic:
                rep = cyclic_boundaries(val, (self.limits[0][self.mask], self.limits[1][self.mask]))
            elif self.limits[0] is None and self.limits[1] is None:
                rep = val
            else:
                rep = boundaries(
                    val,
                    (
                        None if self.limits[0] is None else self.limits[0][self.mask],
                        None if self.limits[1] is None else self.limits[1][self.mask],
                    ),
                )
            return rep

        mask = self.vector_mask()
        flat = self.flat(include_locked=False, include_links=False)

        loc = 0
        reps = []
        for node in flat.values():
            reps.append(
                node.vector_transform_val_to_rep(
                    val[mask[:loc].sum().int() : mask[: loc + node.size].sum().int()]
                )
            )
            loc += node.size
        if len(reps) > 0:
            return torch.cat(reps)
        return torch.tensor((), dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    def _set_val_self(self, val):
        """Handles the setting of the value for a leaf node. Ensures the
        value is a Tensor and that it has the right shape. Will also
        check the limits of the value which has different behaviour
        depending on if it is cyclic, one sided, or two sided.

        """
        val = torch.as_tensor(val, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        if self.shape is not None:
            self._value = val.reshape(self.shape)
        else:
            self._value = val
            self.shape = self._value.shape

        if self.cyclic:
            self._value = self.limits[0] + (
                (self._value - self.limits[0]) % (self.limits[1] - self.limits[0])
            )
            return
        if self.limits[0] is not None:
            if not torch.all(self._value > self.limits[0]):
                raise InvalidParameter(
                    f"{self.name} has lower limit {self.limits[0].detach().cpu().tolist()}"
                )
        if self.limits[1] is not None:
            if not torch.all(self._value < self.limits[1]):
                raise InvalidParameter(
                    f"{self.name} has upper limit {self.limits[1].detach().cpu().tolist()}"
                )

    def _soft_set_val_self(self, val):
        """The same as ``_set_val_self`` except that it doesn't raise an
        error when the values are set outside their range, instead it
        will push the values into the range defined by the limits.

        """
        val = torch.as_tensor(val, dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        if self.shape is not None:
            self._value = val.reshape(self.shape)
        else:
            self._value = val
            self.shape = self._value.shape

        if self.cyclic:
            self._value = self.limits[0] + (
                (self._value - self.limits[0]) % (self.limits[1] - self.limits[0])
            )
            return
        if self.limits[0] is not None:
            self._value = torch.maximum(
                self._value, self.limits[0] + torch.ones_like(self._value) * 1e-3
            )
        if self.limits[1] is not None:
            self._value = torch.minimum(
                self._value, self.limits[1] - torch.ones_like(self._value) * 1e-3
            )

    @value.setter
    def value(self, val):
        if self.locked and not Node.global_unlock:
            return
        if val is None:
            self._value = None
            self.shape = None
            return
        if isinstance(val, str):
            self._value = val
            return
        if isinstance(val, Parameter_Node):
            self._value = val
            self.shape = None
            # Link only to the pointed node
            self.dump()
            self.link(val)
            return
        if isinstance(val, FunctionType):
            self._value = val
            self.shape = None
            return
        if len(self.nodes) > 0:
            self.vector_set_values(val)
            self.shape = None
            return
        self._set_val_self(val)
        self.dump()

    @property
    def shape(self):
        try:
            if isinstance(self._value, Parameter_Node):
                return self._value.shape
            if isinstance(self._value, FunctionType):
                return self.value.shape
            if self.leaf:
                return self._shape
        except AttributeError:
            pass
        return None

    @shape.setter
    def shape(self, shape):
        self._shape = shape

    @property
    def prof(self):
        return self._prof

    @prof.setter
    def prof(self, prof):
        if self.locked and not Node.global_unlock:
            return
        if prof is None:
            self._prof = None
            return
        self._prof = torch.as_tensor(prof, dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    @property
    def uncertainty(self):
        return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, unc):
        if self.locked and not Node.global_unlock:
            return
        if unc is None:
            self._uncertainty = None
            return

        self._uncertainty = torch.as_tensor(
            unc, dtype=AP_config.ap_dtype, device=AP_config.ap_device
        )
        # Ensure that the uncertainty tensor has the same shape as the data
        if self.shape is not None:
            if self._uncertainty.shape != self.shape:
                self._uncertainty = self._uncertainty * torch.ones(
                    self.shape, dtype=AP_config.ap_dtype, device=AP_config.ap_device
                )

    @property
    def limits(self):
        return self._limits

    @limits.setter
    def limits(self, limits):
        if self.locked and not Node.global_unlock:
            return
        if limits[0] is None:
            low = None
        else:
            low = torch.as_tensor(limits[0], dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        if limits[1] is None:
            high = None
        else:
            high = torch.as_tensor(limits[1], dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        self._limits = (low, high)

    def to(self, dtype=None, device=None):
        """
        updates the datatype or device of this parameter
        """
        if dtype is not None:
            dtype = AP_config.ap_dtype
        if device is not None:
            device = AP_config.ap_device

        if isinstance(self._value, torch.Tensor):
            self._value = self._value.to(dtype=dtype, device=device)
        elif len(self.nodes) > 0:
            for node in self.nodes.values():
                node.to(dtype, device)
        if isinstance(self._uncertainty, torch.Tensor):
            self._uncertainty = self._uncertainty.to(dtype=dtype, device=device)
        if isinstance(self.prof, torch.Tensor):
            self.prof = self.prof.to(dtype=dtype, device=device)
        return self

    def get_state(self):
        """Return the values representing the current state of the parameter,
        this can be used to re-load the state later from memory.

        """
        state = super().get_state()
        if self.value is not None:
            if isinstance(self._value, Node):
                state["value"] = "NODE:" + str(self._value.identity)
            elif isinstance(self._value, FunctionType):
                state["value"] = "FUNCTION:" + self._value.__name__
            else:
                state["value"] = self.value.detach().cpu().numpy().tolist()
        if self.shape is not None:
            state["shape"] = list(self.shape)
        if self.units is not None:
            state["units"] = self.units
        if self.uncertainty is not None:
            state["uncertainty"] = self.uncertainty.detach().cpu().numpy().tolist()
        if not (self.limits[0] is None and self.limits[1] is None):
            save_lim = []
            for i in [0, 1]:
                if self.limits[i] is None:
                    save_lim.append(None)
                else:
                    save_lim.append(self.limits[i].detach().cpu().tolist())
            state["limits"] = save_lim
        if self.cyclic:
            state["cyclic"] = self.cyclic
        if self.prof is not None:
            state["prof"] = self.prof.detach().cpu().tolist()

        return state

    def set_state(self, state):
        """Update the state of the parameter given a state variable which
        holds all information about a variable.

        """

        super().set_state(state)
        save_locked = self.locked
        self.locked = False
        self.units = state.get("units", None)
        self.limits = state.get("limits", (None, None))
        self.cyclic = state.get("cyclic", False)
        self.value = state.get("value", None)
        self.uncertainty = state.get("uncertainty", None)
        self.prof = state.get("prof", None)
        self.locked = save_locked

    def flat_detach(self):
        """Due to the system used to track and update values in the DAG, some
        parts of the computational graph used to determine gradients
        may linger after calling .backward on a model using the
        parameters. This function essentially resets all the leaf
        values so that the full computational graph is freed.

        """
        for P in self.flat().values():
            P.value = P.value.detach()
            if P.uncertainty is not None:
                P.uncertainty = P.uncertainty.detach()
            if P.prof is not None:
                P.prof = P.prof.detach()

    @property
    def size(self):
        if self.leaf:
            return self.value.numel()
        return self.vector_values().numel()

    def __len__(self):
        """The number of elements required to fully describe the DAG. This is
        the number of elements in the vector_values tensor.

        """
        return self.size

    def print_params(self, include_locked=True, include_prof=True, include_id=True):
        if self.leaf:
            return (
                f"{self.name}"
                + (f" (id-{self.identity})" if include_id else "")
                + f": {self.value.detach().cpu().tolist()}"
                + (
                    ""
                    if self.uncertainty is None
                    else f" +- {self.uncertainty.detach().cpu().tolist()}"
                )
                + f" [{self.units}]"
                + (
                    ""
                    if self.limits[0] is None and self.limits[1] is None
                    else f", limits: ({None if self.limits[0] is None else self.limits[0].detach().cpu().tolist()}, {None if self.limits[1] is None else self.limits[1].detach().cpu().tolist()})"
                )
                + (", cyclic" if self.cyclic else "")
                + (", locked" if self.locked else "")
                + (
                    f", prof: {self.prof.detach().cpu().tolist()}"
                    if include_prof and self.prof is not None
                    else ""
                )
            )
        elif isinstance(self._value, Parameter_Node):
            return (
                self.name
                + (f" (id-{self.identity})" if include_id else "")
                + " points to: "
                + self._value.print_params(
                    include_locked=include_locked,
                    include_prof=include_prof,
                    include_id=include_id,
                )
            )
        return (
            self.name
            + (
                f" (id-{self.identity}, {('function node, '+self._value.__name__) if isinstance(self._value, FunctionType) else 'branch node'})"
                if include_id
                else ""
            )
            + ":\n"
        )

    def __str__(self):
        reply = self.print_params(include_locked=True, include_prof=False, include_id=False)
        if self.leaf or isinstance(self._value, Parameter_Node):
            return reply
        reply += "\n".join(
            node.print_params(include_locked=True, include_prof=False, include_id=False)
            for node in self.flat(include_locked=True, include_links=False).values()
        )
        return reply

    def __repr__(self, level=0, indent="  "):
        reply = indent * level + self.print_params(
            include_locked=True, include_prof=False, include_id=True
        )
        if self.leaf or isinstance(self._value, Parameter_Node):
            return reply
        reply += "\n".join(
            node.__repr__(level=level + 1, indent=indent) for node in self.nodes.values()
        )
        return reply



import unittest
import astrophot as ap
import torch

from utils import get_astropy_wcs

######################################################################
# Image_Header Objects
######################################################################


class TestImageHeader(unittest.TestCase):
    def test_image_header_creation(self):

        # Minimal input
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            zeropoint=22.5,
            pixelscale=0.2,
        )

        self.assertTrue(torch.all(H.origin == 0), "Origin should be assumed zero if not given")

        # Center
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            pixelscale=0.2,
            center=(10, 10),
        )

        self.assertTrue(
            torch.all(H.origin == 8),
            "Center provided, origin should be adjusted accordingly",
        )

        # Origin
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            pixelscale=0.2,
            origin=(10, 10),
        )

        self.assertTrue(torch.all(H.origin == 10), "Origin provided, origin should be as given")

        # Center radec
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            pixelscale=0.2,
            center_radec=(10, 10),
        )

        self.assertTrue(
            torch.allclose(H.plane_to_world(H.center), torch.ones_like(H.center) * 10),
            "Center_radec provided, center should be as given in world coordinates",
        )

        # Origin radec
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            pixelscale=0.2,
            origin_radec=(10, 10),
        )

        self.assertTrue(
            torch.allclose(H.plane_to_world(H.origin), torch.ones_like(H.center) * 10),
            "Origin_radec provided, origin should be as given in world coordinates",
        )

        # Astropy WCS
        wcs = get_astropy_wcs()
        H = ap.image.Image_Header(
            data_shape=(180, 180),
            wcs=wcs,
        )

        sky_coord = wcs.pixel_to_world(*wcs.wcs.crpix)
        wcs_world = torch.tensor((sky_coord.ra.deg, sky_coord.dec.deg))
        self.assertTrue(
            torch.allclose(
                torch.tensor(
                    wcs.wcs.crpix,
                    dtype=ap.AP_config.ap_dtype,
                    device=ap.AP_config.ap_device,
                ),
                H.world_to_pixel(wcs_world),
            ),
            "Astropy WCS initialization should map crval crpix coordinates",
        )

    def test_image_header_wcs_roundtrip(self):

        wcs = get_astropy_wcs()
        # Minimal input
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            zeropoint=22.5,
            wcs=wcs,
        )

        self.assertTrue(
            torch.allclose(
                H.world_to_plane(H.plane_to_world(torch.zeros_like(H.window.reference_radec))),
                torch.zeros_like(H.window.reference_radec),
            ),
            "WCS world/plane roundtrip should return input value",
        )
        self.assertTrue(
            torch.allclose(
                H.pixel_to_plane(H.plane_to_pixel(torch.zeros_like(H.window.reference_radec))),
                torch.zeros_like(H.window.reference_radec),
            ),
            "WCS pixel/plane roundtrip should return input value",
        )
        self.assertTrue(
            torch.allclose(
                H.world_to_pixel(H.pixel_to_world(torch.zeros_like(H.window.reference_radec))),
                torch.zeros_like(H.window.reference_radec),
                atol=1e-6,
            ),
            "WCS world/pixel roundtrip should return input value",
        )

        self.assertTrue(
            torch.allclose(
                H.pixel_to_plane_delta(
                    H.plane_to_pixel_delta(torch.ones_like(H.window.reference_radec))
                ),
                torch.ones_like(H.window.reference_radec),
            ),
            "WCS pixel/plane delta roundtrip should return input value",
        )

    def test_iamge_header_repr(self):

        wcs = get_astropy_wcs()
        # Minimal input
        H = ap.image.Image_Header(
            data_shape=(20, 20),
            zeropoint=22.5,
            wcs=wcs,
        )

        S = str(H)
        R = repr(H)



# Traditional gradient descent with Adam
from time import time
from typing import Sequence
import torch
import numpy as np

from .base import BaseOptimizer
from .. import AP_config

__all__ = ["Grad"]


class Grad(BaseOptimizer):
    """A gradient descent optimization wrapper for AstroPhot_Model objects.

    The default method is "NAdam", a variant of the Adam optimization algorithm.
    This optimizer uses a combination of gradient descent and Nesterov momentum for faster convergence.
    The optimizer is instantiated with a set of initial parameters and optimization options provided by the user.
    The `fit` method performs the optimization, taking a series of gradient steps until a stopping criteria is met.

    Parameters:
        model (AstroPhot_Model): an AstroPhot_Model object with which to perform optimization.
        initial_state (torch.Tensor, optional): an optional initial state for optimization.
        method (str, optional): the optimization method to use for the update step. Defaults to "NAdam".
        patience (int or None, optional): the number of iterations without improvement before the optimizer will exit early. Defaults to None.
        optim_kwargs (dict, optional): a dictionary of keyword arguments to pass to the pytorch optimizer.

    Attributes:
        model (AstroPhot_Model): the AstroPhot_Model object to optimize.
        current_state (torch.Tensor): the current state of the parameters being optimized.
        iteration (int): the number of iterations performed during the optimization.
        loss_history (list): the history of loss values at each iteration of the optimization.
        lambda_history (list): the history of parameter values at each iteration of the optimization.
        optimizer (torch.optimizer): the PyTorch optimizer object being used.
        patience (int or None): the number of iterations without improvement before the optimizer will exit early.
        method (str): the optimization method being used.
        optim_kwargs (dict): the dictionary of keyword arguments passed to the PyTorch optimizer.


    """

    def __init__(self, model: "AstroPhot_Model", initial_state: Sequence = None, **kwargs) -> None:
        """Initialize the gradient descent optimizer.

        Args:
            - model: instance of the model to be optimized.
            - initial_state: Initial state of the model.
            - patience: (optional) If a positive integer, then stop the optimization if there has been no improvement in the loss for this number of iterations.
            - method: (optional) The name of the optimization method to use. Default is NAdam.
            - optim_kwargs: (optional) Keyword arguments to be passed to the optimizer.
        """

        super().__init__(model, initial_state, **kwargs)

        # set parameters from the user
        self.patience = kwargs.get("patience", None)
        self.method = kwargs.get("method", "NAdam").strip()
        self.optim_kwargs = kwargs.get("optim_kwargs", {})
        self.report_freq = kwargs.get("report_freq", 10)

        # Default learning rate if none given. Equalt to 1 / sqrt(parames)
        if "lr" not in self.optim_kwargs:
            self.optim_kwargs["lr"] = 0.1 / (len(self.current_state) ** (0.5))

        # Instantiates the appropriate pytorch optimizer with the initial state and user provided kwargs
        self.current_state.requires_grad = True
        self.optimizer = getattr(torch.optim, self.method)(
            (self.current_state,), **self.optim_kwargs
        )

    def compute_loss(self) -> torch.Tensor:
        Ym = self.model(parameters=self.current_state, as_representation=True).flatten("data")
        Yt = self.model.target[self.model.window].flatten("data")
        W = (
            self.model.target[self.model.window].flatten("variance")
            if self.model.target.has_variance
            else 1.0
        )
        ndf = len(Yt) - len(self.current_state)
        if self.model.target.has_mask:
            mask = self.model.target[self.model.window].flatten("mask")
            ndf -= torch.sum(mask)
            mask = torch.logical_not(mask)
            loss = torch.sum((Ym[mask] - Yt[mask]) ** 2 / W[mask]) / ndf
        else:
            loss = torch.sum((Ym - Yt) ** 2 / W) / ndf
        return loss

    def step(self) -> None:
        """Take a single gradient step. Take a single gradient step.

        Computes the loss function of the model,
        computes the gradient of the parameters using automatic differentiation,
        and takes a step with the PyTorch optimizer.

        """
        self.iteration += 1

        self.optimizer.zero_grad()
        self.model.parameters.flat_detach()

        loss = self.compute_loss()

        loss.backward()

        self.loss_history.append(loss.detach().cpu().item())
        self.lambda_history.append(np.copy(self.current_state.detach().cpu().numpy()))
        if (
            self.iteration % int(self.max_iter / self.report_freq) == 0
        ) or self.iteration == self.max_iter:
            if self.verbose > 0:
                AP_config.ap_logger.info(f"iter: {self.iteration}, loss: {loss.item()}")
            if self.verbose > 1:
                AP_config.ap_logger.info(f"gradient: {self.current_state.grad}")
        self.optimizer.step()

    def fit(self) -> "BaseOptimizer":
        """
        Perform an iterative fit of the model parameters using the specified optimizer.

        The fit procedure continues until a stopping criteria is met,
        such as the maximum number of iterations being reached,
        or no improvement being made after a specified number of iterations.

        """
        start_fit = time()
        try:
            while True:
                self.step()
                if self.iteration >= self.max_iter:
                    self.message = self.message + " fail max iteration reached"
                    break
                if (
                    self.patience is not None
                    and (len(self.loss_history) - np.argmin(self.loss_history)) > self.patience
                ):
                    self.message = self.message + " fail no improvement"
                    break
                L = np.sort(self.loss_history)
                if len(L) >= 3 and 0 < L[1] - L[0] < 1e-6 and 0 < L[2] - L[1] < 1e-6:
                    self.message = self.message + " success"
                    break
        except KeyboardInterrupt:
            self.message = self.message + " fail interrupted"

        # Set the model parameters to the best values from the fit and clear any previous model sampling
        self.model.parameters.vector_set_representation(self.res())
        if self.verbose > 1:
            AP_config.ap_logger.info(
                f"Grad Fitting complete in {time() - start_fit} sec with message: {self.message}"
            )
        return self



import numpy as np
from scipy.stats import iqr


def Angle_Average(a):
    """
    Compute the average for a list of angles, which may wrap around a cyclic boundary.

    a: list of angles in the range [0,2pi]
    """
    i = np.cos(a) + 1j * np.sin(a)
    return np.angle(np.mean(i))


def Angle_Median(a):
    """
    Compute the median for a list of angles, which may wrap around a cyclic boundary.

    a: list of angles in the range [0,2pi]
    """
    i = np.median(np.cos(a)) + 1j * np.median(np.sin(a))
    return np.angle(i)


def Angle_Scatter(a):
    """
    Compute the scatter for a list of angles, which may wrap around a cyclic boundary.

    a: list of angles in the range [0,2pi]
    """
    i = np.cos(a) + 1j * np.sin(a)
    return iqr(np.angle(1j * i / np.mean(i)), rng=[16, 84])


def Angle_COM_PA(flux, X=None, Y=None):
    """Performs a center of angular mass calculation by using the flux as
    weights to compute a position angle which accounts for the general
    "direction" of the light. This PA is computed mod pi since these
    are 180 degree rotation symmetric.

    Args:
      flux: the weight values for each element (by assumption, pixel fluxes) in a 2D array
      X: x coordinate of the flux points. Assumed centered pixel indices if not given
      Y: y coordinate of the flux points. Assumed centered pixel indices if not given

    """
    if X is None:
        S = flux.shape
        X, Y = np.meshgrid(np.arange(S[1]) - S[1] / 2, np.arange(S[0]) - S[0] / 2, indexing="xy")

    theta = np.arctan2(Y, X)

    ang_com_cos = np.sum(flux * np.cos(2 * theta)) / np.sum(flux)
    ang_com_sin = np.sum(flux * np.sin(2 * theta)) / np.sum(flux)

    return np.arctan2(ang_com_sin, ang_com_cos) / 2 % np.pi



import io
from typing import Optional

import torch
import yaml

from ..utils.conversions.dict_to_hdf5 import dict_to_hdf5, hdf5_to_dict
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..image import Window, Target_Image, Target_Image_List
from ..param import Parameter_Node
from ._shared_methods import select_target, select_sample
from .. import AP_config
from ..errors import NameNotAllowed, InvalidTarget, UnrecognizedModel, InvalidWindow

__all__ = ("AstroPhot_Model",)


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


######################################################################
class AstroPhot_Model(object):
    """Core class for all AstroPhot models and model like objects. This
    class defines the signatures to interact with AstroPhot models
    both for users and internal functions.

    Basic usage:

    .. code-block:: python

      import astrophot as ap

      # Create a model object
      model = ap.models.AstroPhot_Model(
          name="unique name",
          model_type="choose a model type",
          target="Target_Image object",
          window="[[xmin, xmax],[ymin,ymax]]",  # <window pixel coordinates>,
          parameters="dict of parameter specifications if desired",
      )

      # Initialize parameters that weren't set on creation
      model.initialize()

      # Fit model to target
      result = ap.fit.lm(model, verbose=1).fit()

      # Plot the model
      fig, ax = plt.subplots()
      ap.plots.model_image(fig, ax, model)
      plt.show()

      # Sample the model
      img = model()
      pixels = img.data

    AstroPhot models are one of the main ways that one interacts with
    the code, either by setting model parameters or passing models to
    other objects, one can perform a huge variety of fitting
    tasks. The subclass `Component_Model` should be thought of as the
    basic unit when constructing a model of an image while a
    `Group_Model` is a composite structure that may represent a
    complex object, a region of an image, or even a model spanning
    many images. Constructing the `Component_Model`s is where most
    work goes, these store the actual parameters that will be
    optimized. It is important to remember that a `Component_Model`
    only ever applies to a single image and a single component (star,
    galaxy, or even sub-component of one of those) in that image.

    A complex representation is made by stacking many
    `Component_Model`s together, in total this may result in a very
    large number of parameters. Trying to find starting values for all
    of these parameters can be tedious and error prone, so instead all
    built-in AstroPhot models can self initialize and find reasonable
    starting parameters for most situations. Even still one may find
    that for extremely complex fits, it is more stable to first run an
    iterative fitter before global optimization to start the models in
    better initial positions.

    Args:
        name (Optional[str]): every AstroPhot model should have a unique name
        model_type (str): a model type string can determine which kind of AstroPhot model is instantiated.
        target (Optional[Target_Image]): A Target_Image object which stores information about the image which the model is trying to fit.
        filename (Optional[str]): name of a file to load AstroPhot parameters, window, and name. The model will still need to be told its target, device, and other information

    """

    model_type = "model"
    default_uncertainty = 1e-2  # During initialization, uncertainty will be assumed 1% of initial value if no uncertainty is given
    usable = False
    model_names = []

    def __new__(cls, *, filename=None, model_type=None, **kwargs):
        if filename is not None:
            state = AstroPhot_Model.load(filename)
            MODELS = AstroPhot_Model.List_Models()
            for M in MODELS:
                if M.model_type == state["model_type"]:
                    return super(AstroPhot_Model, cls).__new__(M)
            else:
                raise UnrecognizedModel(f"Unknown AstroPhot model type: {state['model_type']}")
        elif model_type is not None:
            MODELS = AstroPhot_Model.List_Models()  # all_subclasses(AstroPhot_Model)
            for M in MODELS:
                if M.model_type == model_type:
                    return super(AstroPhot_Model, cls).__new__(M)
            else:
                raise UnrecognizedModel(f"Unknown AstroPhot model type: {model_type}")

        return super().__new__(cls)

    def __init__(self, *, name=None, target=None, window=None, locked=False, **kwargs):
        if not hasattr(self, "_window"):
            self._window = None
        if not hasattr(self, "_target"):
            self._target = None
        self.name = name
        AP_config.ap_logger.debug("Creating model named: {self.name}")
        self.parameters = Parameter_Node(self.name)
        self.target = target
        self.window = window
        self._locked = locked
        self.mask = kwargs.get("mask", None)

    @property
    def name(self):
        """The name for this model as a string. The name should be unique
        though this is not enforced here. The name should not contain
        the `|` or `:` characters as these are reserved for internal
        use. If one tries to set the name of a model as `None` (for
        example by not providing a name for the model) then a new
        unique name will be generated. The unique name is just the
        model type for this model with an extra unique id appended to
        the end in the format of `[#]` where `#` is a number that
        increases until a unique name is found.

        """
        return self._name

    @name.setter
    def name(self, name):
        try:
            if name == self.name:
                return
        except AttributeError:
            pass
        if name is None:
            i = 0
            while True:
                proposed_name = f"{self.model_type} [{i}]"
                if proposed_name in AstroPhot_Model.model_names:
                    i += 1
                else:
                    name = proposed_name
                    break
        if ":" in name or "|" in name:
            raise NameNotAllowed(
                "characters '|' and ':' are reserved for internal model operations please do not include these in a model name"
            )
        self._name = name
        AstroPhot_Model.model_names.append(name)

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        """When this function finishes, all parameters should have numerical
        values (non None) that are reasonable estimates of the final
        values.

        """
        pass

    def make_model_image(self, window: Optional[Window] = None):
        """This is called to create a blank `Model_Image` object of the
        correct format for this model. This is typically used
        internally to construct the model image before filling the
        pixel values with the model.

        """
        if window is None:
            window = self.window
        else:
            window = self.window & window
        return self.target[window].model_image()

    def sample(self, image=None, window=None, parameters=None, *args, **kwargs):
        """Calling this function should fill the given image with values
        sampled from the given model.

        """
        pass

    def fit_mask(self):
        """
        Return a mask to be used for fitting this model. This will block out
        pixels that are not relevant to the model.
        """
        return torch.zeros_like(self.target[self.window].mask)

    def negative_log_likelihood(
        self,
        parameters=None,
        as_representation=False,
    ):
        """
        Compute the negative log likelihood of the model wrt the target image in the appropriate window.
        """
        if parameters is not None:
            if as_representation:
                self.parameters.vector_set_representation(parameters)
            else:
                self.parameters.vector_set_values(parameters)

        model = self.sample()
        data = self.target[self.window]
        weight = data.weight
        if self.target.has_mask:
            if isinstance(data, Target_Image_List):
                mask = tuple(torch.logical_not(submask) for submask in data.mask)
                chi2 = sum(
                    torch.sum(((mo - da).data ** 2 * wgt)[ma]) / 2.0
                    for mo, da, wgt, ma in zip(model, data, weight, mask)
                )
            else:
                mask = torch.logical_not(data.mask)
                chi2 = torch.sum(((model - data).data ** 2 * weight)[mask]) / 2.0
        else:
            if isinstance(data, Target_Image_List):
                chi2 = sum(
                    torch.sum(((mo - da).data ** 2 * wgt)) / 2.0
                    for mo, da, wgt in zip(model, data, weight)
                )
            else:
                chi2 = torch.sum(((model - data).data ** 2 * weight)) / 2.0

        return chi2

    def jacobian(
        self,
        parameters=None,
        **kwargs,
    ):
        raise NotImplementedError("please use a subclass of AstroPhot_Model")

    @default_internal
    def total_flux(self, parameters=None, window=None, image=None):
        F = self(parameters=parameters, window=None, image=None)
        return torch.sum(F.data)

    @property
    def window(self):
        """The window defines a region on the sky in which this model will be
        optimized and typically evaluated. Two models with
        non-overlapping windows are in effect independent of each
        other. If there is another model with a window that spans both
        of them, then they are tenuously connected.

        If not provided, the model will assume a window equal to the
        target it is fitting. Note that in this case the window is not
        explicitly set to the target window, so if the model is moved
        to another target then the fitting window will also change.

        """
        if self._window is None:
            if self.target is None:
                raise ValueError(
                    "This model has no target or window, these must be provided by the user"
                )
            return self.target.window.copy()
        return self._window

    def set_window(self, window):
        if window is None:
            # If no window given, set to none
            self._window = None
        elif isinstance(window, Window):
            # If window object given, use that
            self._window = window
        elif len(window) == 2:
            # If window given in pixels, use relative to target
            self._window = self.target.window.copy().crop_to_pixel(window)
        else:
            raise InvalidWindow(f"Unrecognized window format: {str(window)}")

    @window.setter
    def window(self, window):
        self.set_window(window)

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, tar):
        if not (tar is None or isinstance(tar, Target_Image)):
            raise InvalidTarget("AstroPhot_Model target must be a Target_Image instance.")
        self._target = tar

    @property
    def locked(self):
        """Set when the model should remain fixed going forward. This model
        will be bypassed when fitting parameters, however it will
        still be sampled for generating the model image.

        Warning:

          This feature is not yet fully functional and should be avoided for now. It is included here for the sake of testing.

        """
        return self._locked

    @locked.setter
    def locked(self, val):
        self._locked = val

    @property
    def parameter_order(self):
        """Returns the model parameters in the order they are kept for
        flattening, such as when evaluating the model with a tensor of
        parameter values.

        """
        return tuple(P.name for P in self.parameters)

    def __str__(self):
        """String representation for the model."""
        return self.parameters.__str__()

    def __repr__(self):
        """Detailed string representation for the model."""
        return yaml.dump(self.get_state(), indent=2)

    def get_state(self, *args, **kwargs):
        """Returns a dictionary of the state of the model with its name,
        type, parameters, and other important information. This
        dictionary is what gets saved when a model saves to disk.

        """
        state = {
            "name": self.name,
            "model_type": self.model_type,
        }
        return state

    def save(self, filename="AstroPhot.yaml"):
        """Saves a model object to disk. By default the file type should be
        yaml, this is the only file type which gets tested, though
        other file types such as json and hdf5 should work.

        """
        if filename.endswith(".yaml"):
            state = self.get_state()
            with open(filename, "w") as f:
                yaml.dump(state, f, indent=2)
        elif filename.endswith(".json"):
            import json

            state = self.get_state()
            with open(filename, "w") as f:
                json.dump(state, f, indent=2)
        elif filename.endswith(".hdf5"):
            import h5py

            state = self.get_state()
            with h5py.File(filename, "w") as F:
                dict_to_hdf5(F, state)
        else:
            if isinstance(filename, str) and "." in filename:
                raise ValueError(
                    f"Unrecognized filename format: {filename[filename.find('.'):]}, must be one of: .json, .yaml, .hdf5"
                )
            else:
                raise ValueError(
                    f"Unrecognized filename format: {str(filename)}, must be one of: .json, .yaml, .hdf5"
                )

    @classmethod
    def load(cls, filename="AstroPhot.yaml"):
        """
        Loads a saved model object.
        """
        if isinstance(filename, dict):
            state = filename
        elif isinstance(filename, io.TextIOBase):
            state = yaml.load(filename, Loader=yaml.FullLoader)
        elif filename.endswith(".yaml"):
            with open(filename, "r") as f:
                state = yaml.load(f, Loader=yaml.FullLoader)
        elif filename.endswith(".json"):
            import json

            with open(filename, "r") as f:
                state = json.load(f)
        elif filename.endswith(".hdf5"):
            import h5py

            with h5py.File(filename, "r") as F:
                state = hdf5_to_dict(F)
        else:
            if isinstance(filename, str) and "." in filename:
                raise ValueError(
                    f"Unrecognized filename format: {filename[filename.find('.'):]}, must be one of: .json, .yaml, .hdf5"
                )
            else:
                raise ValueError(
                    f"Unrecognized filename format: {str(filename)}, must be one of: .json, .yaml, .hdf5 or python dictionary."
                )
        return state

    @classmethod
    def List_Models(cls, usable=None):
        MODELS = all_subclasses(cls)
        if usable is not None:
            for model in list(MODELS):
                if model.usable is not usable:
                    MODELS.remove(model)
        return MODELS

    @classmethod
    def List_Model_Names(cls, usable=None):
        MODELS = cls.List_Models(usable=usable)
        names = []
        for model in MODELS:
            names.append(model.model_type)
        return list(sorted(names, key=lambda n: n[::-1]))

    def __eq__(self, other):
        return self is other

    def __getitem__(self, key):
        return self.parameters[key]

    def __contains__(self, key):
        return self.parameters.__contains__(key)

    def __del__(self):
        try:
            i = AstroPhot_Model.model_names.index(self.name)
            AstroPhot_Model.model_names.pop(i)
        except:
            pass

    @select_sample
    def __call__(
        self,
        image=None,
        parameters=None,
        window=None,
        as_representation=False,
        **kwargs,
    ):

        if parameters is None:
            parameters = self.parameters
        elif isinstance(parameters, torch.Tensor):
            if as_representation:
                self.parameters.vector_set_representation(parameters)
            else:
                self.parameters.vector_set_values(parameters)
            parameters = self.parameters
        return self.sample(image=image, window=window, parameters=parameters, **kwargs)



from functools import wraps
import inspect
import warnings

import numpy as np


def ignore_numpy_warnings(func):
    """This decorator is used to turn off numpy warnings. This should
    only be used in initialize scripts which often run heuristic code
    to determine initial parameter values. These heuristics may
    encounter log(0) or sqrt(-1) or other numerical artifacts and
    should handle them before returning. This decorator simply cleans
    up that processes to minimize clutter in the output.

    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        old_settings = np.seterr(all="ignore")
        warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
        result = func(*args, **kwargs)
        np.seterr(**old_settings)
        warnings.filterwarnings("default", category=np.VisibleDeprecationWarning)
        return result

    return wrapped


def default_internal(func):
    """This decorator inspects the input parameters for a function which
    expects to receive `image` and `parameters` arguments. If either
    of these are not given, then the model can use its default values
    for the parameters assuming the `image` is the internal `target`
    object and the `parameters` are the internally stored parameters.

    """
    sig = inspect.signature(func)

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        bound = sig.bind(self, *args, **kwargs)
        bound.apply_defaults()

        if bound.arguments.get("image") is None:
            bound.arguments["image"] = self.target
        if bound.arguments.get("parameters") is None:
            bound.arguments["parameters"] = self.parameters

        return func(*bound.args, **bound.kwargs)

    return wrapper



import numpy as np
import torch

from .galaxy_model_object import Galaxy_Model
from ..utils.interpolate import cubic_spline_torch
from ..utils.conversions.coordinates import Rotate_Cartesian
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..param import Param_Unlock, Param_SoftLimits
from ._shared_methods import select_target

__all__ = ["Warp_Galaxy"]


class Warp_Galaxy(Galaxy_Model):
    """Galaxy model which includes radially varrying PA and q
    profiles. This works by warping the coordinates using the same
    transform for a global PA/q except applied to each pixel
    individually. In the limit that PA and q are a constant, this
    recovers a basic galaxy model with global PA/q. However, a linear
    PA profile will give a spiral appearance, variations of PA/q
    profiles can create complex galaxy models. The form of the
    coordinate transformation looks like:

    X, Y = meshgrid(image)
    R = sqrt(X^2 + Y^2)
    X', Y' = Rot(theta(R), X, Y)
    Y'' = Y' / q(R)

    where the definitions are the same as for a regular galaxy model,
    except now the theta is a function of radius R (before
    transformation) and the axis ratio q is also a function of radius
    (before the transformation).

    Parameters:
        q(R): Tensor of axis ratio values for axis ratio spline
        PA(R): Tensor of position angle values as input to the spline

    """

    model_type = f"warp {Galaxy_Model.model_type}"
    parameter_specs = {
        "q(R)": {"units": "b/a", "limits": (0.05, 1), "uncertainty": 0.04},
        "PA(R)": {
            "units": "rad",
            "limits": (0, np.pi),
            "cyclic": True,
            "uncertainty": 0.08,
        },
    }
    _parameter_order = Galaxy_Model._parameter_order + ("q(R)", "PA(R)")
    usable = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        # create the PA(R) and q(R) profile radii if needed
        for prof_param in ["PA(R)", "q(R)"]:
            if parameters[prof_param].prof is None:
                if parameters[prof_param].value is None:  # from scratch
                    new_prof = [0, 2 * target.pixel_length]
                    while new_prof[-1] < torch.min(self.window.shape / 2):
                        new_prof.append(
                            new_prof[-1] + torch.max(2 * target.pixel_length, new_prof[-1] * 0.2)
                        )
                    new_prof.pop()
                    new_prof.pop()
                    new_prof.append(torch.sqrt(torch.sum((self.window.shape / 2) ** 2)))
                    parameters[prof_param].prof = new_prof
                else:  # matching length of a provided profile
                    # create logarithmically spaced profile radii
                    new_prof = [0] + list(
                        np.logspace(
                            np.log10(2 * target.pixel_length),
                            np.log10(torch.max(self.window.shape / 2).item()),
                            len(parameters[prof_param].value) - 1,
                        )
                    )
                    # ensure no step is smaller than a pixelscale
                    for i in range(1, len(new_prof)):
                        if new_prof[i] - new_prof[i - 1] < target.pixel_length.item():
                            new_prof[i] = new_prof[i - 1] + target.pixel_length.item()
                    parameters[prof_param].prof = new_prof

        if not (parameters["PA(R)"].value is None or parameters["q(R)"].value is None):
            return

        with Param_Unlock(parameters["PA(R)"]), Param_SoftLimits(parameters["PA(R)"]):
            if parameters["PA(R)"].value is None:
                parameters["PA(R)"].value = np.zeros(len(parameters["PA(R)"].prof)) + target.north
            if parameters["PA(R)"].uncertainty is None:
                parameters["PA(R)"].uncertainty = (5 * np.pi / 180) * torch.ones_like(
                    parameters["PA(R)"].value
                )
            if parameters["q(R)"].value is None:
                # If no initial value is provided for q(R) a heuristic initial value is assumed.
                # The most neutral initial position would be 1, but the boundaries of q are (0,1) non-inclusive
                # so that is not allowed. A value like 0.999 may get stuck since it is near the very edge of
                # the (0,1) range. So 0.9 is chosen to be mostly passive, but still some signal for the optimizer.
                parameters["q(R)"].value = np.ones(len(parameters["q(R)"].prof)) * 0.9
            if parameters["q(R)"].uncertainty is None:
                parameters["q(R)"].uncertainty = self.default_uncertainty * parameters["q(R)"].value

    @default_internal
    def transform_coordinates(self, X, Y, image=None, parameters=None):
        X, Y = super().transform_coordinates(X, Y, image, parameters)
        R = self.radius_metric(X, Y, image, parameters)
        PA = cubic_spline_torch(
            parameters["PA(R)"].prof,
            -(parameters["PA(R)"].value - image.north),
            R.view(-1),
        ).view(*R.shape)
        q = cubic_spline_torch(parameters["q(R)"].prof, parameters["q(R)"].value, R.view(-1)).view(
            *R.shape
        )
        X, Y = Rotate_Cartesian(PA, X, Y)
        return X, Y / q



from typing import Optional

import torch

from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from .ray_model import Ray_Galaxy
from .psf_model_object import PSF_Model
from .superellipse_model import SuperEllipse_Galaxy, SuperEllipse_Warp
from .foureirellipse_model import FourierEllipse_Galaxy, FourierEllipse_Warp
from .wedge_model import Wedge_Galaxy
from ._shared_methods import (
    parametric_initialize,
    parametric_segment_initialize,
    select_target,
)
from ..param import Parameter_Node
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.parametric_profiles import exponential_np

__all__ = [
    "Exponential_Galaxy",
    "Exponential_PSF",
    "Exponential_SuperEllipse",
    "Exponential_SuperEllipse_Warp",
    "Exponential_Warp",
    "Exponential_Ray",
    "Exponential_Wedge",
]


def _x0_func(model_params, R, F):
    return R[4], F[4]


def _wrap_exp(R, re, ie):
    return exponential_np(R, re, 10**ie)


class Exponential_Galaxy(Galaxy_Model):
    """basic galaxy model with a exponential profile for the radial light
    profile. The light profile is defined as:

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {Galaxy_Model.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters: Optional[Parameter_Node] = None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model


class Exponential_PSF(PSF_Model):
    """basic point source model with a exponential profile for the radial light
    profile.

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {PSF_Model.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)", "value": 0.0, "locked": True},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = PSF_Model._parameter_order + ("Re", "Ie")
    usable = True
    model_integrated = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model
    from ._shared_methods import radial_evaluate_model as evaluate_model


class Exponential_SuperEllipse(SuperEllipse_Galaxy):
    """super ellipse galaxy model with a exponential profile for the radial
    light profile.

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {SuperEllipse_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = SuperEllipse_Galaxy._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model


class Exponential_SuperEllipse_Warp(SuperEllipse_Warp):
    """super ellipse warp galaxy model with a exponential profile for the
    radial light profile.

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {SuperEllipse_Warp.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = SuperEllipse_Warp._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model


class Exponential_FourierEllipse(FourierEllipse_Galaxy):
    """fourier mode perturbations to ellipse galaxy model with an
    exponential profile for the radial light profile.

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {FourierEllipse_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = FourierEllipse_Galaxy._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model


class Exponential_FourierEllipse_Warp(FourierEllipse_Warp):
    """fourier mode perturbations to ellipse galaxy model with a exponential
    profile for the radial light profile.

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {FourierEllipse_Warp.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = FourierEllipse_Warp._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model


class Exponential_Warp(Warp_Galaxy):
    """warped coordinate galaxy model with a exponential profile for the
    radial light model.

    I(R) = Ie * exp(-b1(R/Re - 1))

    where I(R) is the brightness as a function of semi-major axis, Ie
    is the brightness at the half light radius, b1 is a constant not
    involved in the fit, R is the semi-major axis, and Re is the
    effective radius.

    Parameters:
        Ie: Brightness at half light radius, represented as the log of the brightness divided by pixelscale squared. This is proportional to a surface brightness
        Re: half light radius, represented in arcsec. This parameter cannot go below zero.

    """

    model_type = f"exponential {Warp_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Warp_Galaxy._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_exp, ("Re", "Ie"), _x0_func)

    from ._shared_methods import exponential_radial_model as radial_model


class Exponential_Ray(Ray_Galaxy):
    """ray galaxy model with a sersic profile for the radial light
    model. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius.

    Parameters:
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"exponential {Ray_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Ray_Galaxy._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            parameters=parameters,
            target=target,
            prof_func=_wrap_exp,
            params=("Re", "Ie"),
            x0_func=_x0_func,
            segments=self.rays,
        )

    from ._shared_methods import exponential_iradial_model as iradial_model


class Exponential_Wedge(Wedge_Galaxy):
    """wedge galaxy model with a exponential profile for the radial light
    model. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius.

    Parameters:
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"exponential {Wedge_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Wedge_Galaxy._parameter_order + ("Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            parameters=parameters,
            target=target,
            prof_func=_wrap_exp,
            params=("Re", "Ie"),
            x0_func=_x0_func,
            segments=self.wedges,
        )

    from ._shared_methods import exponential_iradial_model as iradial_model



import sys
import os
import importlib
import numpy as np
from astropy.io import fits
from ..image import Target_Image
from ..models import AstroPhot_Model
from ..fit import LM
from .. import AP_config

__all__ = ["basic_config"]


def GetOptions(c):
    newoptions = {}
    for var in dir(c):
        if var.startswith("ap_"):
            val = getattr(c, var)
            if val is not None:
                newoptions[var] = val
    return newoptions


def import_configfile(config_file):
    if "/" in config_file:
        startat = config_file.rfind("/") + 1
    else:
        startat = 0
    if "." in config_file:
        use_config = config_file[startat : config_file.rfind(".")]
    else:
        use_config = config_file[startat:]
    if startat > 0:
        sys.path.append(os.path.abspath(config_file[: config_file.rfind("/")]))
    else:
        sys.path.append(os.getcwd())
    c = importlib.import_module(use_config)
    return c


def basic_config(config_file):
    c = import_configfile(config_file)  # importlib.import_module(config_file)
    config = GetOptions(c)

    # Parse Target
    ######################################################################
    AP_config.ap_logger.info("Collecting target information")
    target = config.get("ap_target", None)
    if target is None:
        target_file = config.get("ap_target_file", None)
        target_hdu = config.get("ap_target_hdu", 0)
        variance_file = config.get("ap_variance_file", None)
        variance_hdu = config.get("ap_variance_hdu", 0)
        target_pixelscale = config.get("ap_target_pixelscale", None)
        target_zeropoint = config.get("ap_target.zeropoint", None)
        target_origin = config.get("ap_target_origin", None)

        if variance_file is not None:
            var_data = np.array(fits.open(target_file)[target_hdu].data, dtype=np.float64)
        else:
            var_data = None
        if target_file is not None:
            data = np.array(fits.open(target_file)[target_hdu].data, dtype=np.float64)
            target = Target_Image(
                data=data,
                pixelscale=target_pixelscale,
                zeropoint=target_zeropoint,
                variance=var_data,
                origin=target_origin,
            )

    # Parse Models
    ######################################################################
    AP_config.ap_logger.info("Constructing models")
    model_info_list = config.get("ap_models", [])
    name_order = config.get(
        "ap_model_name_order",
        list(n[9:] for n in filter(lambda k: k.startswith("ap_model_"), config.keys())),
    )
    for name in name_order:
        key_name = "ap_model_" + name
        model_info_list.append(config[key_name])
        if "name" not in model_info_list[-1]:
            model_info_list[-1]["name"] = name
    model_list = []
    for model in model_info_list:
        model_list.append(AstroPhot_Model(target=target, **model))

    MODEL = AstroPhot_Model(
        name="AstroPhot",
        model_type="group model",
        models=model_list,
        target=target,
    )

    # Parse Optimize
    ######################################################################
    AP_config.ap_logger.info("Running optimization")
    MODEL.initialize()

    optim_type = config.get("ap_optimizer", "LM")
    optim_kwargs = config.get("ap_optimizer_kwargs", {})
    if optim_type is None:
        # perform no optimization, simply write the astrophot model and the requested images
        pass
    elif optim_type == "LM":
        result = LM(MODEL, **optim_kwargs).fit()

    # Parse Save
    ######################################################################
    AP_config.ap_logger.info("Saving model")
    model_save = config.get("ap_saveto_model", "AstroPhot.yaml")
    MODEL.save(model_save)

    model_image_save = config.get("ap_saveto_model_image", None)
    if model_image_save is not None:
        MODEL().save(model_image_save)

    model_residual_save = config.get("ap_saveto_model_residual", None)
    if model_residual_save is not None:
        (target - MODEL()).save(model_residual_save)



from typing import Optional

import torch
import numpy as np
from scipy.stats import iqr

from ..utils.initialize import isophotes
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.angle_operations import Angle_COM_PA
from ..utils.conversions.coordinates import (
    Rotate_Cartesian,
)
from ..param import Param_Unlock, Param_SoftLimits, Parameter_Node
from .model_object import Component_Model
from ._shared_methods import select_target


__all__ = ["Galaxy_Model"]


class Galaxy_Model(Component_Model):
    """General galaxy model to be subclassed for any specific
    representation. Defines a galaxy as an object with a position
    angle and axis ratio, or effectively a tilted disk. Most
    subclassing models should simply define a radial model or update
    to the coordinate transform. The definition of the position angle and axis ratio used here is simply a scaling along the minor axis. The transformation can be written as:

    X, Y = meshgrid(image)
    X', Y' = Rot(theta, X, Y)
    Y'' = Y' / q

    where X Y are the coordinates of an image, X' Y' are the rotated
    coordinates, Rot is a rotation matrix by angle theta applied to the
    initial X Y coordinates, Y'' is the scaled semi-minor axis, and q
    is the axis ratio.

    Parameters:
        q: axis ratio to scale minor axis from the ratio of the minor/major axis b/a, this parameter is unitless, it is restricted to the range (0,1)
        PA: position angle of the smei-major axis relative to the image positive x-axis in radians, it is a cyclic parameter in the range [0,pi)

    """

    model_type = f"galaxy {Component_Model.model_type}"
    parameter_specs = {
        "q": {"units": "b/a", "limits": (0, 1), "uncertainty": 0.03},
        "PA": {
            "units": "radians",
            "limits": (0, np.pi),
            "cyclic": True,
            "uncertainty": 0.06,
        },
    }
    _parameter_order = Component_Model._parameter_order + ("q", "PA")
    usable = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters: Optional[Parameter_Node] = None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        if not (parameters["PA"].value is None or parameters["q"].value is None):
            return
        target_area = target[self.window]
        target_dat = target_area.data.detach().cpu().numpy()
        if target_area.has_mask:
            mask = target_area.mask.detach().cpu().numpy()
            target_dat[mask] = np.median(target_dat[np.logical_not(mask)])
        edge = np.concatenate(
            (
                target_dat[:, 0],
                target_dat[:, -1],
                target_dat[0, :],
                target_dat[-1, :],
            )
        )
        edge_average = np.nanmedian(edge)
        edge_scatter = iqr(edge[np.isfinite(edge)], rng=(16, 84)) / 2
        icenter = target_area.plane_to_pixel(parameters["center"].value)

        if parameters["PA"].value is None:
            weights = target_dat - edge_average
            Coords = target_area.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
            X, Y = X.detach().cpu().numpy(), Y.detach().cpu().numpy()
            if target_area.has_mask:
                seg = np.logical_not(target_area.mask.detach().cpu().numpy())
                PA = Angle_COM_PA(weights[seg], X[seg], Y[seg])
            else:
                PA = Angle_COM_PA(weights, X, Y)

            with Param_Unlock(parameters["PA"]), Param_SoftLimits(parameters["PA"]):
                parameters["PA"].value = (PA + target_area.north) % np.pi
                if parameters["PA"].uncertainty is None:
                    parameters["PA"].uncertainty = (5 * np.pi / 180) * torch.ones_like(
                        parameters["PA"].value
                    )  # default uncertainty of 5 degrees is assumed
        if parameters["q"].value is None:
            q_samples = np.linspace(0.2, 0.9, 15)
            iso_info = isophotes(
                target_area.data.detach().cpu().numpy() - edge_average,
                (icenter[1].detach().cpu().item(), icenter[0].detach().cpu().item()),
                threshold=3 * edge_scatter,
                pa=(parameters["PA"].value - target.north).detach().cpu().item(),
                q=q_samples,
            )
            with Param_Unlock(parameters["q"]), Param_SoftLimits(parameters["q"]):
                parameters["q"].value = q_samples[
                    np.argmin(list(iso["amplitude2"] for iso in iso_info))
                ]
                if parameters["q"].uncertainty is None:
                    parameters["q"].uncertainty = parameters["q"].value * self.default_uncertainty

    @default_internal
    def transform_coordinates(self, X, Y, image=None, parameters=None):
        X, Y = Rotate_Cartesian(-(parameters["PA"].value - image.north), X, Y)
        return (
            X,
            Y / parameters["q"].value,
        )

    @default_internal
    def evaluate_model(
        self, X=None, Y=None, image=None, parameters: Parameter_Node = None, **kwargs
    ):
        if X is None or Y is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        XX, YY = self.transform_coordinates(X, Y, image, parameters)
        return self.radial_model(
            self.radius_metric(XX, YY, image, parameters),
            image=image,
            parameters=parameters,
        )



import torch

from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from .ray_model import Ray_Galaxy
from .wedge_model import Wedge_Galaxy
from .psf_model_object import PSF_Model
from .superellipse_model import SuperEllipse_Galaxy, SuperEllipse_Warp
from .foureirellipse_model import FourierEllipse_Galaxy, FourierEllipse_Warp
from ._shared_methods import (
    parametric_initialize,
    parametric_segment_initialize,
    select_target,
)
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.parametric_profiles import sersic_np
from ..utils.conversions.functions import (
    sersic_Ie_to_flux_torch,
    general_uncertainty_prop,
)


__all__ = [
    "Sersic_Galaxy",
    "Sersic_PSF",
    "Sersic_Warp",
    "Sersic_SuperEllipse",
    "Sersic_FourierEllipse",
    "Sersic_Ray",
    "Sersic_Wedge",
    "Sersic_SuperEllipse_Warp",
    "Sersic_FourierEllipse_Warp",
]


def _x0_func(model, R, F):
    return 2.0, R[4], F[4]


def _wrap_sersic(R, n, r, i):
    return sersic_np(R, n, r, 10 ** (i))


class Sersic_Galaxy(Galaxy_Model):
    """basic galaxy model with a sersic profile for the radial light
    profile. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {Galaxy_Model.model_type}"
    parameter_specs = {
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
        "Ie": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    @default_internal
    def total_flux(self, parameters=None):
        return sersic_Ie_to_flux_torch(
            10 ** parameters["Ie"].value,
            parameters["n"].value,
            parameters["Re"].value,
            parameters["q"].value,
        )

    @default_internal
    def total_flux_uncertainty(self, parameters=None):
        return general_uncertainty_prop(
            (
                10 ** parameters["Ie"].value,
                parameters["n"].value,
                parameters["Re"].value,
                parameters["q"].value,
            ),
            (
                (10 ** parameters["Ie"].value)
                * parameters["Ie"].uncertainty
                * torch.log(10 * torch.ones_like(parameters["Ie"].value)),
                parameters["n"].uncertainty,
                parameters["Re"].uncertainty,
                parameters["q"].uncertainty,
            ),
            sersic_Ie_to_flux_torch,
        )

    def _integrate_reference(self, image_data, image_header, parameters):
        tot = self.total_flux(parameters)
        return tot / image_data.numel()

    from ._shared_methods import sersic_radial_model as radial_model


class Sersic_PSF(PSF_Model):
    """basic point source model with a sersic profile for the radial light
    profile. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {PSF_Model.model_type}"
    parameter_specs = {
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
        "Ie": {
            "units": "log10(flux/arcsec^2)",
            "value": 0.0,
            "uncertainty": 0.0,
            "locked": True,
        },
    }
    _parameter_order = PSF_Model._parameter_order + ("n", "Re", "Ie")
    usable = True
    model_integrated = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    from ._shared_methods import sersic_radial_model as radial_model
    from ._shared_methods import radial_evaluate_model as evaluate_model


class Sersic_SuperEllipse(SuperEllipse_Galaxy):
    """super ellipse galaxy model with a sersic profile for the radial
    light profile. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {SuperEllipse_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = SuperEllipse_Galaxy._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    from ._shared_methods import sersic_radial_model as radial_model


class Sersic_SuperEllipse_Warp(SuperEllipse_Warp):
    """super ellipse warp galaxy model with a sersic profile for the
    radial light profile. The functional form of the Sersic profile is
    defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {SuperEllipse_Warp.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = SuperEllipse_Warp._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    from ._shared_methods import sersic_radial_model as radial_model


class Sersic_FourierEllipse(FourierEllipse_Galaxy):
    """fourier mode perturbations to ellipse galaxy model with a sersic
    profile for the radial light profile. The functional form of the
    Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {FourierEllipse_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = FourierEllipse_Galaxy._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    from ._shared_methods import sersic_radial_model as radial_model


class Sersic_FourierEllipse_Warp(FourierEllipse_Warp):
    """fourier mode perturbations to ellipse galaxy model with a sersic
    profile for the radial light profile. The functional form of the
    Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {FourierEllipse_Warp.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = FourierEllipse_Warp._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    from ._shared_methods import sersic_radial_model as radial_model


class Sersic_Warp(Warp_Galaxy):
    """warped coordinate galaxy model with a sersic profile for the radial
    light model. The functional form of the Sersic profile is defined
    as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {Warp_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Warp_Galaxy._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(self, parameters, target, _wrap_sersic, ("n", "Re", "Ie"), _x0_func)

    from ._shared_methods import sersic_radial_model as radial_model


class Sersic_Ray(Ray_Galaxy):
    """ray galaxy model with a sersic profile for the radial light
    model. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {Ray_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Ray_Galaxy._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            target=target,
            parameters=parameters,
            prof_func=_wrap_sersic,
            params=("n", "Re", "Ie"),
            x0_func=_x0_func,
            segments=self.rays,
        )

    from ._shared_methods import sersic_iradial_model as iradial_model


class Sersic_Wedge(Wedge_Galaxy):
    """wedge galaxy model with a sersic profile for the radial light
    model. The functional form of the Sersic profile is defined as:

    I(R) = Ie * exp(- bn((R/Re)^(1/n) - 1))

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, Ie is the brightness as the
    half light radius, bn is a function of n and is not involved in
    the fit, Re is the half light radius, and n is the sersic index
    which controls the shape of the profile.

    Parameters:
        n: Sersic index which controls the shape of the brightness profile
        Ie: brightness at the half light radius, represented as the log of the brightness divided by pixel scale squared.
        Re: half light radius

    """

    model_type = f"sersic {Wedge_Galaxy.model_type}"
    parameter_specs = {
        "Ie": {"units": "log10(flux/arcsec^2)"},
        "n": {"units": "none", "limits": (0.36, 8), "uncertainty": 0.05},
        "Re": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Wedge_Galaxy._parameter_order + ("n", "Re", "Ie")
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_segment_initialize(
            model=self,
            parameters=parameters,
            target=target,
            prof_func=_wrap_sersic,
            params=("n", "Re", "Ie"),
            x0_func=_x0_func,
            segments=self.wedges,
        )

    from ._shared_methods import sersic_iradial_model as iradial_model



import torch
import numpy as np


def Rotate_Cartesian(theta, X, Y=None):
    """
    Applies a rotation matrix to the X,Y coordinates
    """
    s = torch.sin(theta)
    c = torch.cos(theta)
    if Y is None:
        return c * X[0] - s * X[1], s * X[0] + c * X[1]
    return c * X - s * Y, s * X + c * Y


def Rotate_Cartesian_np(theta, X, Y):
    """
    Applies a rotation matrix to the X,Y coordinates
    """
    s = np.sin(theta)
    c = np.cos(theta)
    return c * X - s * Y, c * Y + s * X


def Axis_Ratio_Cartesian(q, X, Y, theta=0.0, inv_scale=False):
    """
    Applies the transformation: R(theta) Q R(-theta)
    where R is the rotation matrix and Q is the matrix which scales the y component by 1/q.
    This effectively counter-rotates the coordinates so that the angle theta is along the x-axis
    then applies the y-axis scaling, then re-rotates everything back to where it was.
    """
    if inv_scale:
        scale = (1 / q) - 1
    else:
        scale = q - 1
    ss = 1 + scale * torch.pow(torch.sin(theta), 2)
    cc = 1 + scale * torch.pow(torch.cos(theta), 2)
    s2 = scale * torch.sin(2 * theta)
    return ss * X - s2 * Y / 2, -s2 * X / 2 + cc * Y


def Axis_Ratio_Cartesian_np(q, X, Y, theta=0.0, inv_scale=False):
    """
    Applies the transformation: R(theta) Q R(-theta)
    where R is the rotation matrix and Q is the matrix which scales the y component by 1/q.
    This effectively counter-rotates the coordinates so that the angle theta is along the x-axis
    then applies the y-axis scaling, then re-rotates everything back to where it was.
    """
    if inv_scale:
        scale = (1 / q) - 1
    else:
        scale = q - 1
    ss = 1 + scale * np.sin(theta) ** 2
    cc = 1 + scale * np.cos(theta) ** 2
    s2 = scale * np.sin(2 * theta)
    return ss * X - s2 * Y / 2, -s2 * X / 2 + cc * Y



import torch
import numpy as np

from ..utils.decorators import ignore_numpy_warnings, default_internal
from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from ._shared_methods import select_target
from ..param import Param_Unlock, Param_SoftLimits
from .. import AP_config

__all__ = ["FourierEllipse_Galaxy", "FourierEllipse_Warp"]


class FourierEllipse_Galaxy(Galaxy_Model):
    """Expanded galaxy model which includes a Fourier transformation in
    its radius metric. This allows for the expression of arbitrarily
    complex isophotes instead of pure ellipses. This is a common
    extension of the standard elliptical representation. The form of
    the Fourier perturbations is:

    R' = R * exp(sum_m(a_m * cos(m * theta + phi_m)))

    where R' is the new radius value, R is the original ellipse
    radius, a_m is the amplitude of the m'th Fourier mode, m is the
    index of the Fourier mode, theta is the angle around the ellipse,
    and phi_m is the phase of the m'th fourier mode. This
    representation is somewhat different from other Fourier mode
    implementations where instead of an expoenntial it is just 1 +
    sum_m(...), we opt for this formulation as it is more numerically
    stable. It cannot ever produce negative radii, but to first order
    the two representation are the same as can be seen by a Taylor
    expansion of exp(x) = 1 + x + O(x^2).

    One can create extremely complex shapes using different Fourier
    modes, however usually it is only low order modes that are of
    interest. For intuition, the first Fourier mode is roughly
    equivalent to a lopsided galaxy, one side will be compressed and
    the opposite side will be expanded. The second mode is almost
    never used as it is nearly degenerate with ellipticity. The third
    mode is an alternate kind of lopsidedness for a galaxy which makes
    it somewhat triangular, meaning that it is wider on one side than
    the other. The fourth mode is similar to a boxyness/diskyness
    parameter which tends to make more pronounced peanut shapes since
    it is more rounded than a superellipse representation. Modes
    higher than 4 are only useful in very specialized situations. In
    general one should consider carefully why the Fourier modes are
    being used for the science case at hand.

    Parameters:
        am: Tensor of amplitudes for the Fourier modes, indicates the strength of each mode.
        phi_m: Tensor of phases for the Fourier modes, adjusts the orientation of the mode perturbation relative to the major axis. It is cyclically defined in the range [0,2pi)

    """

    model_type = f"fourier {Galaxy_Model.model_type}"
    parameter_specs = {
        "am": {"units": "none"},
        "phim": {"units": "radians", "limits": (0, 2 * np.pi), "cyclic": True},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("am", "phim")
    modes = (1, 3, 4)
    track_attrs = Galaxy_Model.track_attrs + ["modes"]
    usable = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modes = torch.tensor(
            kwargs.get("modes", FourierEllipse_Galaxy.modes),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
        )

    @default_internal
    def angular_metric(self, X, Y, image=None, parameters=None):
        return torch.atan2(Y, X)

    @default_internal
    def radius_metric(self, X, Y, image=None, parameters=None):
        R = super().radius_metric(X, Y, image, parameters)
        theta = self.angular_metric(X, Y, image, parameters)
        return R * torch.exp(
            torch.sum(
                parameters["am"].value.view(len(self.modes), -1)
                * torch.cos(
                    self.modes.view(len(self.modes), -1) * theta.view(-1)
                    + parameters["phim"].value.view(len(self.modes), -1)
                ),
                0,
            ).view(theta.shape)
        )

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        with Param_Unlock(parameters["am"]), Param_SoftLimits(parameters["am"]):
            if parameters["am"].value is None:
                parameters["am"].value = torch.zeros(
                    len(self.modes),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
            if parameters["am"].uncertainty is None:
                parameters["am"].uncertainty = torch.tensor(
                    self.default_uncertainty * np.ones(len(self.modes)),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
        with Param_Unlock(parameters["phim"]), Param_SoftLimits(parameters["phim"]):
            if parameters["phim"].value is None:
                parameters["phim"].value = torch.zeros(
                    len(self.modes),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
            if parameters["phim"].uncertainty is None:
                parameters["phim"].uncertainty = (
                    torch.tensor(  # Uncertainty assumed to be 5 degrees if not provided
                        (5 * np.pi / 180) * np.ones(len(self.modes)),
                        dtype=AP_config.ap_dtype,
                        device=AP_config.ap_device,
                    )
                )


class FourierEllipse_Warp(Warp_Galaxy):
    """Expanded warp galaxy model which includes a Fourier transformation
    in its radius metric. This allows for the expression of
    arbitrarily complex isophotes instead of pure ellipses. This is a
    common extension of the standard elliptical representation. The
    form of the Fourier perturbations is:

    R' = R * exp(sum_m(a_m * cos(m * theta + phi_m)))

    where R' is the new radius value, R is the original ellipse
    radius, a_m is the amplitude of the m'th Fourier mode, m is the
    index of the Fourier mode, theta is the angle around the ellipse,
    and phi_m is the phase of the m'th fourier mode. This
    representation is somewhat different from other Fourier mode
    implementations where instead of an expoenntial it is just 1 +
    sum_m(...), we opt for this formulation as it is more numerically
    stable. It cannot ever produce negative radii, but to first order
    the two representation are the same as can be seen by a Taylor
    expansion of exp(x) = 1 + x + O(x^2).

    One can create extremely complex shapes using different Fourier
    modes, however usually it is only low order modes that are of
    interest. For intuition, the first Fourier mode is roughly
    equivalent to a lopsided galaxy, one side will be compressed and
    the opposite side will be expanded. The second mode is almost
    never used as it is nearly degenerate with ellipticity. The third
    mode is an alternate kind of lopsidedness for a galaxy which makes
    it somewhat triangular, meaning that it is wider on one side than
    the other. The fourth mode is similar to a boxyness/diskyness
    parameter which tends to make more pronounced peanut shapes since
    it is more rounded than a superellipse representation. Modes
    higher than 4 are only useful in very specialized situations. In
    general one should consider carefully why the Fourier modes are
    being used for the science case at hand.

    Parameters:
        am: Tensor of amplitudes for the Fourier modes, indicates the strength of each mode.
        phi_m: Tensor of phases for the Fourier modes, adjusts the orientation of the mode perturbation relative to the major axis. It is cyclically defined in the range [0,2pi)

    """

    model_type = f"fourier {Warp_Galaxy.model_type}"
    parameter_specs = {
        "am": {"units": "none"},
        "phim": {"units": "radians", "limits": (0, 2 * np.pi), "cyclic": True},
    }
    _parameter_order = Warp_Galaxy._parameter_order + ("am", "phim")
    modes = (1, 3, 4)
    track_attrs = Galaxy_Model.track_attrs + ["modes"]
    usable = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modes = torch.tensor(
            kwargs.get("modes", FourierEllipse_Warp.modes),
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
        )

    @default_internal
    def angular_metric(self, X, Y, image=None, parameters=None):
        return torch.atan2(Y, X)

    @default_internal
    def radius_metric(self, X, Y, image=None, parameters=None):
        R = super().radius_metric(X, Y, image, parameters)
        theta = self.angular_metric(X, Y, image, parameters)
        return R * torch.exp(
            torch.sum(
                parameters["am"].value.view(len(self.modes), -1)
                * torch.cos(
                    self.modes.view(len(self.modes), -1) * theta.view(-1)
                    + parameters["phim"].value.view(len(self.modes), -1)
                ),
                0,
            ).view(theta.shape)
        )

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        with Param_Unlock(parameters["am"]), Param_SoftLimits(parameters["am"]):
            if parameters["am"].value is None:
                parameters["am"].value = torch.zeros(
                    len(self.modes),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
            if parameters["am"].uncertainty is None:
                parameters["am"].uncertainty = torch.tensor(
                    self.default_uncertainty * np.ones(len(self.modes)),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
        with Param_Unlock(parameters["phim"]), Param_SoftLimits(parameters["phim"]):
            if parameters["phim"].value is None:
                parameters["phim"].value = torch.zeros(
                    len(self.modes),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
            if parameters["phim"].uncertainty is None:
                parameters["phim"].uncertainty = torch.tensor(
                    (5 * np.pi / 180)
                    * np.ones(
                        len(self.modes)
                    ),  # Uncertainty assumed to be 5 degrees if not provided
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )



from typing import Optional, Union, Any, Sequence, Tuple

import torch
from torch.nn.functional import pad
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS as AstropyWCS

from .window_object import Window, Window_List
from .image_header import Image_Header
from .. import AP_config
from ..errors import SpecificationConflict, ConflicingWCS, InvalidData, InvalidWindow

__all__ = ["Image", "Image_List"]


class Image(object):
    """Core class to represent images with pixel values, pixel scale,
       and a window defining the spatial coordinates on the sky.
       It supports arithmetic operations with other image objects while preserving logical image boundaries.
       It also provides methods for determining the coordinate locations of pixels

    Parameters:
        data: the matrix of pixel values for the image
        pixelscale: the length of one side of a pixel in arcsec/pixel
        window: an AstroPhot Window object which defines the spatial coordinates on the sky
        filename: a filename from which to load the image.
        zeropoint: photometric zero point for converting from pixel flux to magnitude
        metadata: Any information the user wishes to associate with this image, stored in a python dictionary
        origin: The origin of the image in the coordinate system.
    """

    def __init__(
        self,
        *,
        data: Optional[torch.Tensor] = None,
        header: Optional[Image_Header] = None,
        wcs: Optional[AstropyWCS] = None,
        pixelscale: Optional[Union[float, torch.Tensor]] = None,
        window: Optional[Window] = None,
        filename: Optional[str] = None,
        zeropoint: Optional[Union[float, torch.Tensor]] = None,
        metadata: Optional[dict] = None,
        origin: Optional[Sequence] = None,
        center: Optional[Sequence] = None,
        identity: str = None,
        state: Optional[dict] = None,
        fits_state: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize an instance of the APImage class.

        Parameters:
        -----------
        data : numpy.ndarray or None, optional
            The image data. Default is None.
        wcs : astropy.wcs.wcs.WCS or None, optional
            A WCS object which defines a coordinate system for the image. Note that AstroPhot only handles basic WCS conventions. It will use the WCS object to get `wcs.pixel_to_world(-0.5, -0.5)` to determine the position of the origin in world coordinates. It will also extract the `pixel_scale_matrix` to index pixels going forward.
        pixelscale : float or None, optional
            The physical scale of the pixels in the image, in units of arcseconds. Default is None.
        window : Window or None, optional
            A Window object defining the area of the image to use. Default is None.
        filename : str or None, optional
            The name of a file containing the image data. Default is None.
        zeropoint : float or None, optional
            The image's zeropoint, used for flux calibration. Default is None.
        metadata : dict or None, optional
            Any information the user wishes to associate with this image, stored in a python dictionary. Default is None.
        origin : numpy.ndarray or None, optional
            The origin of the image in the coordinate system, as a 1D array of length 2. Default is None.
        center : numpy.ndarray or None, optional
            The center of the image in the coordinate system, as a 1D array of length 2. Default is None.

        Returns:
        --------
        None
        """
        self._data = None

        if state is not None:
            self.header = Image_Header(state=state["header"])
        elif fits_state is not None:
            self.set_fits_state(fits_state)
            return
        elif header is None:
            if data is None and window is None and filename is None:
                raise InvalidData("Image must have either data or a window to construct itself.")
            self.header = Image_Header(
                data_shape=None if data is None else data.shape,
                pixelscale=pixelscale,
                wcs=wcs,
                window=window,
                filename=filename,
                zeropoint=zeropoint,
                metadata=metadata,
                origin=origin,
                center=center,
                identity=identity,
                **kwargs,
            )
        else:
            self.header = header

        if filename is not None:
            self.load(filename)
        elif state is not None:
            self.set_state(state)
        elif fits_state is not None:
            self.data = fits_state[0]["DATA"]
        else:
            # set the data
            if data is None:
                self.data = torch.zeros(
                    torch.flip(self.window.pixel_shape, (0,)).detach().cpu().tolist(),
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
            else:
                self.data = data

            self.to()

        # # Check that image data and header are in agreement (this requires talk back from GPU to CPU so is only used for testing)
        # assert np.all(np.flip(np.array(self.data.shape)[:2]) == self.window.pixel_shape.numpy()), f"data shape {np.flip(np.array(self.data.shape)[:2])}, window shape {self.window.pixel_shape.numpy()}"

    @property
    def north(self):
        return self.header.north

    @property
    def pixel_area(self):
        return self.header.pixel_area

    @property
    def pixel_length(self):
        return self.header.pixel_length

    def world_to_plane(self, *args, **kwargs):
        return self.window.world_to_plane(*args, **kwargs)

    def plane_to_world(self, *args, **kwargs):
        return self.window.plane_to_world(*args, **kwargs)

    def plane_to_pixel(self, *args, **kwargs):
        return self.window.plane_to_pixel(*args, **kwargs)

    def pixel_to_plane(self, *args, **kwargs):
        return self.window.pixel_to_plane(*args, **kwargs)

    def plane_to_pixel_delta(self, *args, **kwargs):
        return self.window.plane_to_pixel_delta(*args, **kwargs)

    def pixel_to_plane_delta(self, *args, **kwargs):
        return self.window.pixel_to_plane_delta(*args, **kwargs)

    def world_to_pixel(self, *args, **kwargs):
        return self.window.world_to_pixel(*args, **kwargs)

    def pixel_to_world(self, *args, **kwargs):
        return self.window.pixel_to_world(*args, **kwargs)

    def get_coordinate_meshgrid(self):
        return self.window.get_coordinate_meshgrid()

    def get_coordinate_corner_meshgrid(self):
        return self.window.get_coordinate_corner_meshgrid()

    def get_coordinate_simps_meshgrid(self):
        return self.window.get_coordinate_simps_meshgrid()

    @property
    def origin(self) -> torch.Tensor:
        """
        Returns the origin (bottom-left corner) of the image window.

        Returns:
            torch.Tensor: A 1D tensor of shape (2,) containing the (x, y) coordinates of the origin.
        """
        return self.header.window.origin

    @property
    def shape(self) -> torch.Tensor:
        """
        Returns the shape (size) of the image window.

        Returns:
                torch.Tensor: A 1D tensor of shape (2,) containing the (width, height) of the window in pixels.
        """
        return self.header.window.shape

    @property
    def center(self) -> torch.Tensor:
        """
        Returns the center of the image window.

        Returns:
            torch.Tensor: A 1D tensor of shape (2,) containing the (x, y) coordinates of the center.
        """
        return self.header.window.center

    @property
    def size(self) -> torch.Tensor:
        """
        Returns the size of the image window, the number of pixels in the image.

        Returns:
            torch.Tensor: A 0D tensor containing the number of pixels.
        """
        return self.header.window.size

    @property
    def window(self):
        return self.header.window

    @property
    def pixelscale(self):
        return self.header.pixelscale

    @property
    def zeropoint(self):
        return self.header.zeropoint

    @property
    def metadata(self):
        return self.header.metadata

    @property
    def identity(self):
        return self.header.identity

    @property
    def data(self) -> torch.Tensor:
        """
        Returns the image data.
        """
        return self._data

    @data.setter
    def data(self, data) -> None:
        """Set the image data."""
        self.set_data(data)

    def set_data(self, data: Union[torch.Tensor, np.ndarray], require_shape: bool = True):
        """
        Set the image data.

        Args:
            data (torch.Tensor or numpy.ndarray): The image data.
            require_shape (bool): Whether to check that the shape of the data is the same as the current data.

        Raises:
            SpecificationConflict: If `require_shape` is `True` and the shape of the data is different from the current data.
        """
        if self._data is not None and require_shape and data.shape != self._data.shape:
            raise SpecificationConflict(
                f"Attempting to change image data with tensor that has a different shape! ({data.shape} vs {self._data.shape}) Use 'require_shape = False' if this is desired behaviour."
            )

        if data is None:
            self.data = torch.tensor((), dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        elif isinstance(data, torch.Tensor):
            self._data = data.to(dtype=AP_config.ap_dtype, device=AP_config.ap_device)
        else:
            self._data = torch.as_tensor(data, dtype=AP_config.ap_dtype, device=AP_config.ap_device)

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        return self.__class__(
            data=torch.clone(self.data),
            header=self.header.copy(**kwargs),
            **kwargs,
        )

    def blank_copy(self, **kwargs):
        """Produces a blank copy of the image which has the same properties
        except that its data is now filled with zeros.

        """
        return self.__class__(
            data=torch.zeros_like(self.data),
            header=self.header.copy(**kwargs),
            **kwargs,
        )

    def get_window(self, window, **kwargs):
        """Get a sub-region of the image as defined by a window on the sky."""
        return self.__class__(
            data=self.data[self.window.get_self_indices(window)],
            header=self.header.get_window(window, **kwargs),
            **kwargs,
        )

    def to(self, dtype=None, device=None):
        if dtype is None:
            dtype = AP_config.ap_dtype
        if device is None:
            device = AP_config.ap_device
        if self._data is not None:
            self._data = self._data.to(dtype=dtype, device=device)
        self.header.to(dtype=dtype, device=device)
        return self

    def crop(self, pixels):
        # does this show up?
        if len(pixels) == 1:  # same crop in all dimension
            self.set_data(
                self.data[
                    pixels[0].int() : (self.data.shape[0] - pixels[0]).int(),
                    pixels[0].int() : (self.data.shape[1] - pixels[0]).int(),
                ],
                require_shape=False,
            )
        elif len(pixels) == 2:  # different crop in each dimension
            self.set_data(
                self.data[
                    pixels[1].int() : (self.data.shape[0] - pixels[1]).int(),
                    pixels[0].int() : (self.data.shape[1] - pixels[0]).int(),
                ],
                require_shape=False,
            )
        elif len(pixels) == 4:  # different crop on all sides
            self.set_data(
                self.data[
                    pixels[2].int() : (self.data.shape[0] - pixels[3]).int(),
                    pixels[0].int() : (self.data.shape[1] - pixels[1]).int(),
                ],
                require_shape=False,
            )
        self.header = self.header.crop(pixels)
        return self

    def flatten(self, attribute: str = "data") -> np.ndarray:
        return getattr(self, attribute).reshape(-1)

    def get_coordinate_meshgrid(self):
        return self.header.get_coordinate_meshgrid()

    def get_coordinate_corner_meshgrid(self):
        return self.header.get_coordinate_corner_meshgrid()

    def get_coordinate_simps_meshgrid(self):
        return self.header.get_coordinate_simps_meshgrid()

    def reduce(self, scale: int, **kwargs):
        """This operation will downsample an image by the factor given. If
        scale = 2 then 2x2 blocks of pixels will be summed together to
        form individual larger pixels. A new image object will be
        returned with the appropriate pixelscale and data tensor. Note
        that the window does not change in this operation since the
        pixels are condensed, but the pixel size is increased
        correspondingly.

        Parameters:
            scale: factor by which to condense the image pixels. Each scale X scale region will be summed [int]

        """
        if not isinstance(scale, int) and not (
            isinstance(scale, torch.Tensor) and scale.dtype is torch.int32
        ):
            raise SpecificationConflict(f"Reduce scale must be an integer! not {type(scale)}")
        if scale == 1:
            return self

        MS = self.data.shape[0] // scale
        NS = self.data.shape[1] // scale
        return self.__class__(
            data=self.data[: MS * scale, : NS * scale]
            .reshape(MS, scale, NS, scale)
            .sum(axis=(1, 3)),
            header=self.header.rescale_pixel(scale, **kwargs),
            **kwargs,
        )

    def expand(self, padding: Tuple[float]) -> None:
        """
        Args:
          padding tuple[float]: length 4 tuple with amounts to pad each dimension in physical units
        """
        padding = np.array(padding)
        if np.any(padding < 0):
            raise SpecificationConflict("negative padding not allowed in expand method")
        pad_boundaries = tuple(np.int64(np.round(np.array(padding) / self.pixelscale)))
        self.data = pad(self.data, pad=pad_boundaries, mode="constant", value=0)
        self.header.expand(padding)

    def get_state(self):
        state = {}
        state["type"] = self.__class__.__name__
        state["data"] = self.data.detach().cpu().tolist()
        state["header"] = self.header.get_state()
        return state

    def set_state(self, state):
        self.set_data(state["data"], require_shape=False)
        self.header.set_state(state["header"])

    def get_fits_state(self):
        states = [{}]
        states[0]["DATA"] = self.data.detach().cpu().numpy()
        states[0]["HEADER"] = self.header.get_fits_state()
        states[0]["HEADER"]["IMAGE"] = "PRIMARY"
        return states

    def set_fits_state(self, states):
        for state in states:
            if state["HEADER"]["IMAGE"] == "PRIMARY":
                self.set_data(np.array(state["DATA"], dtype=np.float64), require_shape=False)
                self.header.set_fits_state(state["HEADER"])
                break

    def save(self, filename=None, overwrite=True):
        states = self.get_fits_state()
        img_list = [fits.PrimaryHDU(states[0]["DATA"], header=fits.Header(states[0]["HEADER"]))]
        for state in states[1:]:
            img_list.append(fits.ImageHDU(state["DATA"], header=fits.Header(state["HEADER"])))
        hdul = fits.HDUList(img_list)
        if filename is not None:
            hdul.writeto(filename, overwrite=overwrite)
        return hdul

    def load(self, filename):
        hdul = fits.open(filename)
        states = list({"DATA": hdu.data, "HEADER": hdu.header} for hdu in hdul)
        self.set_fits_state(states)

    def __sub__(self, other):
        if isinstance(other, Image):
            new_img = self[other.window].copy()
            new_img.data -= other.data[self.window.get_other_indices(other)]
            return new_img
        else:
            new_img = self.copy()
            new_img.data -= other
            return new_img

    def __add__(self, other):
        if isinstance(other, Image):
            new_img = self[other.window].copy()
            new_img.data += other.data[self.window.get_other_indices(other)]
            return new_img
        else:
            new_img = self.copy()
            new_img.data += other
            return new_img

    def __iadd__(self, other):
        if isinstance(other, Image):
            self.data[other.window.get_other_indices(self)] += other.data[
                self.window.get_other_indices(other)
            ]
        else:
            self.data += other
        return self

    def __isub__(self, other):
        if isinstance(other, Image):
            self.data[other.window.get_other_indices(self)] -= other.data[
                self.window.get_other_indices(other)
            ]
        else:
            self.data -= other
        return self

    def __getitem__(self, *args):
        if len(args) == 1 and isinstance(args[0], Window):
            return self.get_window(args[0])
        if len(args) == 1 and isinstance(args[0], Image):
            return self.get_window(args[0].window)
        raise ValueError("Unrecognized Image getitem request!")

    def __str__(self):
        return f"image pixelscale: {self.pixelscale.detach().cpu().numpy()} origin: {self.origin.detach().cpu().numpy()} shape: {self.shape.detach().cpu().numpy()}"

    def __repr__(self):
        return f"image pixelscale: {self.pixelscale.detach().cpu().numpy()} origin: {self.origin.detach().cpu().numpy()} shape: {self.shape.detach().cpu().numpy()} center: {self.center.detach().cpu().numpy()}\ndata: {self.data.detach().cpu().numpy()}"


class Image_List(Image):
    def __init__(self, image_list, window=None):
        self.image_list = list(image_list)
        self.check_wcs()
        self.window = window

    def check_wcs(self):
        """Ensure the WCS systems being used by all the windows in this list
        are consistent with each other. They should all project world
        coordinates onto the same tangent plane.

        """
        ref = torch.stack(tuple(I.window.reference_radec for I in self.image_list))
        if not torch.allclose(ref, ref[0]):
            raise ConflicingWCS(
                "Reference (world) coordinate mismatch! All images in Image_List are not on the same tangent plane! Likely serious coordinate mismatch problems. See the coordinates page in the documentation for what this means."
            )
        ref = torch.stack(tuple(I.window.reference_planexy for I in self.image_list))
        if not torch.allclose(ref, ref[0]):
            raise ConflicingWCS(
                "Reference (tangent plane) coordinate mismatch! All images in Image_List are not on the same tangent plane! Likely serious coordinate mismatch problems. See the coordinates page in the documentation for what this means."
            )

        if len(set(I.window.projection for I in self.image_list)) > 1:
            raise ConflicingWCS(
                "Projection mismatch! All images in Image_List are not on the same tangent plane! Likely serious coordinate mismatch problems. See the coordinates page in the documentation for what this means."
            )

    @property
    def window(self):
        return Window_List(list(image.window for image in self.image_list))

    @window.setter
    def window(self, window):
        if window is None:
            return

        if not isinstance(window, Window_List):
            raise InvalidWindow("Target_List must take a Window_List object as its window")

        for i in range(len(self.image_list)):
            self.image_list[i] = self.image_list[i][window.window_list[i]]

    @property
    def pixelscale(self):
        return tuple(image.pixelscale for image in self.image_list)

    @property
    def zeropoint(self):
        return tuple(image.zeropoint for image in self.image_list)

    @property
    def data(self):
        return tuple(image.data for image in self.image_list)

    @data.setter
    def data(self, data):
        for image, dat in zip(self.image_list, data):
            image.data = dat

    def copy(self):
        return self.__class__(
            tuple(image.copy() for image in self.image_list),
        )

    def blank_copy(self):
        return self.__class__(
            tuple(image.blank_copy() for image in self.image_list),
        )

    def get_window(self, window):
        return self.__class__(
            tuple(image[win] for image, win in zip(self.image_list, window)),
        )

    def index(self, other):
        if isinstance(other, Image) and hasattr(other, "identity"):
            for i, self_image in enumerate(self.image_list):
                if other.identity == self_image.identity:
                    return i
            else:
                raise ValueError("Could not find identity match between image list and input image")
        raise NotImplementedError(f"Image_List cannot get index for {type(other)}")

    def to(self, dtype=None, device=None):
        if dtype is not None:
            dtype = AP_config.ap_dtype
        if device is not None:
            device = AP_config.ap_device
        for image in self.image_list:
            image.to(dtype=dtype, device=device)
        return self

    def crop(self, *pixels):
        raise NotImplementedError("Crop function not available for Image_List object")

    def get_coordinate_meshgrid(self):
        return tuple(image.get_coordinate_meshgrid() for image in self.image_list)

    def get_coordinate_corner_meshgrid(self):
        return tuple(image.get_coordinate_corner_meshgrid() for image in self.image_list)

    def get_coordinate_simps_meshgrid(self):
        return tuple(image.get_coordinate_simps_meshgrid() for image in self.image_list)

    def flatten(self, attribute="data"):
        return torch.cat(tuple(image.flatten(attribute) for image in self.image_list))

    def reduce(self, scale):
        if scale == 1:
            return self

        return self.__class__(
            tuple(image.reduce(scale) for image in self.image_list),
        )

    def __sub__(self, other):
        if isinstance(other, Image_List):
            new_list = []
            for self_image, other_image in zip(self.image_list, other.image_list):
                new_list.append(self_image - other_image)
            return self.__class__(new_list)
        else:
            new_list = []
            for self_image, other_image in zip(self.image_list, other):
                new_list.append(self_image - other_image)
            return self.__class__(new_list)

    def __add__(self, other):
        if isinstance(other, Image_List):
            new_list = []
            for self_image, other_image in zip(self.image_list, other.image_list):
                new_list.append(self_image + other_image)
            return self.__class__(new_list)
        else:
            new_list = []
            for self_image, other_image in zip(self.image_list, other):
                new_list.append(self_image + other_image)
            return self.__class__(new_list)

    def __isub__(self, other):
        if isinstance(other, Image_List):
            for self_image, other_image in zip(self.image_list, other.image_list):
                self_image -= other_image
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image -= other_image
        return self

    def __iadd__(self, other):
        if isinstance(other, Image_List):
            for self_image, other_image in zip(self.image_list, other.image_list):
                self_image += other_image
        else:
            for self_image, other_image in zip(self.image_list, other):
                self_image += other_image
        return self

    def save(self, filename=None, overwrite=True):
        raise NotImplementedError("Save/load not yet available for image lists")

    def load(self, filename):
        raise NotImplementedError("Save/load not yet available for image lists")

    def __getitem__(self, *args):
        if len(args) == 1 and isinstance(args[0], Window):
            return self.get_window(args[0])
        if len(args) == 1 and isinstance(args[0], Image):
            return self.get_window(args[0].window)
        if all(isinstance(arg, (int, slice)) for arg in args):
            return self.image_list.__getitem__(*args)
        raise ValueError("Unrecognized Image_List getitem request!")

    def __str__(self):
        return "image list of:\n" + "\n".join(image.__str__() for image in self.image_list)

    def __repr__(self):
        return "image list of:\n" + "\n".join(image.__repr__() for image in self.image_list)

    def __iter__(self):
        return (img for img in self.image_list)

    #     self._index = 0
    #     return self

    # def __next__(self):
    #     if self._index >= len(self.image_list):
    #         raise StopIteration
    #     img = self.image_list[self._index]
    #     self._index += 1
    #     return img
